##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from __future__ import annotations
import logging
from typing import TYPE_CHECKING

from azure.quantum.aio.storage import (
    ContainerClient,
    download_blob,
    BlobClient
)
from azure.quantum.aio.job.job import Job
from azure.quantum.optimization import Problem as SyncProblem
from azure.quantum.optimization import ProblemType

logger = logging.getLogger(__name__)

__all__ = ["Problem", "ProblemType"]

if TYPE_CHECKING:
    from azure.quantum.aio.workspace import Workspace


class Problem(SyncProblem):
    """Problem to submit to the service.

    :param name: Problem name
    :type name: str
    :param terms: Problem terms, depending on solver.
        Defaults to None
    :type terms: Optional[List[Term]], optional
    :param init_config: Optional configuration details, depending on solver.
        Defaults to None
    :type init_config: Optional[Dict[str,int]], optional
    :param problem_type: Problem type (ProblemType.pubo or
        ProblemType.ising), defaults to ProblemType.ising
    :type problem_type: ProblemType, optional
    """
    async def upload(
        self,
        workspace: "Workspace",
        container_name: str = "optimization-problems",
        blob_name: str = "inputData",
        container_uri: str = None
    ):
        """Uploads an optimization problem instance to
        the cloud storage linked with the Workspace.

        :param workspace: interaction terms of the problem.
        :type workspace: Workspace
        :param container_name: Container name, defaults to "optimization-problems"
        :type container_name: str, optional
        :param blob_name: Blob name, defaults to None
        :type blob_name: str, optional
        :param container_uri: Optional container URI
        :type container_uri: str
        :return: uri of the uploaded problem
        :rtype: str
        """
        blob_params = [workspace, container_name, blob_name]
        if self.uploaded_blob_uri and self.uploaded_blob_params == blob_params:
            return self.uploaded_blob_uri

        if blob_name is None:
            blob_name = self._blob_name()

        encoding = "gzip"
        blob = self.to_blob()
        if container_uri is None:
            container_uri = await workspace.get_container_uri(
                container_name=container_name
            )
        input_data_uri = await Job.upload_input_data(
            input_data=blob,
            blob_name=blob_name,
            container_uri=container_uri,
            encoding=encoding,
            content_type="application/json"
        )
        self.uploaded_blob_params = blob_params
        self.uploaded_blob_uri = input_data_uri
        return input_data_uri

    async def download(self, workspace: "Workspace"):
        """Downloads the uploaded problem as an instance of `Problem`"""
        if not self.uploaded_blob_uri:
            raise Exception(
                "Problem may not be downloaded before it is uploaded"
            )
        blob_client = BlobClient.from_blob_url(self.uploaded_blob_uri)
        container_client = ContainerClient.from_container_url(
            await workspace._get_linked_storage_sas_uri(
                blob_client.container_name
            )
        )
        blob_name = blob_client.blob_name
        blob = container_client.get_blob_client(blob_name)
        contents = await download_blob(blob.url)
        return Problem.deserialize(contents, self.name)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from __future__ import annotations
import numpy as np
from typing import List, Dict, Union, Optional
from enum import Enum
from abc import ABC

__all__ = ["TermBase", "Term"]

try:
    import numpy.typing as npt

    WArray = Union[int, float, npt.ArrayLike]

    def _convert_if_numpy_type(param: WArray):
        # Attempt first a conversion to a supported
        # type if parameter is a numpy float/int.
        numpy_integer_types = [
            np.byte,
            np.ubyte,
            np.short,
            np.ushort,
            np.intc,
            np.uintc,
            np.int_,
            np.uint,
            np.longlong,
            np.ulonglong,
            np.int8,
            np.uint8,
            np.int16,
            np.uint16,
            np.int32,
            np.uint32,
            np.int64,
            np.uint64,
        ]

        numpy_float_types = [
            np.float16,
            np.float32,
            np.float64,
            np.float_,
            np.half,
            np.single,
            np.double,
        ]

        if hasattr(param, "__iter__"):
            # Handle scalar-like arrays, if specified.
            param = param[0]

        if (
            hasattr(param, "dtype")
            and param.dtype in numpy_integer_types + numpy_float_types
        ):
            return param.item()
        else:
            return param


except ImportError:
    npt = None
    WArray = Union[int, float]

    def _convert_if_numpy_type(param: WArray):
        return param


class TermBase(ABC):
    """
    Term base class; this class is not directly initialized
    """
    def __init__(
        self,
        c: Optional[WArray] = None,
    ):
        if c is not None:
            # Current intended specification of term.
            coeff = c
        else:
            raise RuntimeError("Cost should be provided for each term.")

        coeff = _convert_if_numpy_type(coeff)
        if type(coeff) != int and type(coeff) != float:
            raise RuntimeError(
                "c must be a float or int value, \
                    or a NumPy value that can be converted to those."
            )
        self.c = coeff

    def to_dict(self):
        return self.__dict__

    @classmethod
    def from_dict(cls, obj):
        return cls(c=obj.get("c"))

    def evaluate(self, *args, **kwargs) -> float:
        """Given a variable configuration, evaluate the value of the term.
        :param configuration:
            The dictionary of variable ids to their assigned value
        """
        return self.c

    def reduce_by_variable_state(self, *args, **kwargs) -> Optional[TermBase]:
        """Given some fixed variable states,
            transform the existing term into new term.
        Returns None if the new term is effectively 0
        :param fixed_variables:
            The dictionary of variable ids and their fixed state
        """
        return TermBase(c=self.c) if self.c != 0 else None

    def __repr__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

class Term(TermBase):
    """
    Class describing a single (monomial) term.
    """
    def __init__(
        self,
        indices: List[int] = None,
        w: Optional[WArray] = None,
        c: Optional[WArray] = None,
    ):
        if w is not None:
            # Legacy support if 'w' is used to specify
            # term instead of the expected 'c'.
            coeff = w
            parameter_name_used = "w"
        elif c is not None:
            coeff = c
            parameter_name_used = "c"
        else:
            raise RuntimeError("Cost should be provided for each term.")
        coeff = _convert_if_numpy_type(coeff)
        if type(coeff) != int and type(coeff) != float:
            raise RuntimeError(
                f"{parameter_name_used} must be a float or int value, \
                    or a NumPy value that can be converted to those."
            )
        TermBase.__init__(self, c=coeff)
        self.ids = indices
    
    @classmethod
    def from_dict(cls, obj: dict):
        return cls(indices=obj.get("ids"), c=obj.get("c"))
    
    def evaluate(self, configuration: Dict[int, int]) -> float:
        """Given a variable configuration, evaluate the value of the term.
        :param configuration:
            The dictionary of variable ids to their assigned value
        """
        try:
            multiplier = (
                np.prod([configuration[i] for i in self.ids])
                if len(self.ids) > 0
                else 1.0
            )
        except KeyError:
            print(
                "Error - variable id found in term {0}, \
                    but not found in the supplied configuration.".format(
                    self.ids
                )
            )
            raise

        return multiplier * self.c
    
    def reduce_by_variable_state(
        self, fixed_variables: Dict[int, int]
    ) -> Optional[Term]:
        """Given some fixed variable states,
            transform the existing term into new term.
        Returns None if the new term is effectively 0
        :param fixed_variables:
            The dictionary of variable ids and their fixed state
        """
        new_ids = []
        new_c = self.c

        for i in self.ids:
            if i not in fixed_variables:
                new_ids.append(i)
            else:
                new_c *= fixed_variables[i]
                if new_c == 0:
                    return None

        return Term(indices=new_ids, c=new_c)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

import asyncio
import logging
import json
import time

from azure.quantum.aio.job.base_job import BaseJob, DEFAULT_TIMEOUT
from azure.quantum.job.job import Job as SyncJob
from azure.quantum.job.filtered_job import FilteredJob

__all__ = ["Job"]

logger = logging.getLogger(__name__)

_log = logging.getLogger(__name__)


class Job(BaseJob, SyncJob, FilteredJob):
    """Azure Quantum Job that is submitted to a given Workspace.

    :param workspace: Workspace instance to submit job to
    :type workspace: Workspace
    :param job_details: Job details model,
            contains Job ID, name and other details
    :type job_details: JobDetails
    """
    async def submit(self):
        """Submit a job to Azure Quantum."""
        _log.debug(f"Submitting job with ID {self.id}")
        job = await self.workspace.submit_job(self)
        self.details = job.details

    async def refresh(self):
        """Refreshes the Job's details by querying the workspace."""
        self.details = (await self.workspace.get_job(self.id)).details

    async def wait_until_completed(
        self,
        max_poll_wait_secs=30,
        timeout_secs=None,
        print_progress=True
    ) -> None:
        """Keeps refreshing the Job's details
        until it reaches a finished status.

        :param max_poll_wait_secs: Maximum poll wait time, defaults to 30
        :type max_poll_wait_secs: int, optional
        :param timeout_secs: Timeout in seconds, defaults to None
        :type timeout_secs: int, optional
        :param print_progress: Print "." to stdout to display progress
        :type print_progress: bool, optional
        :raises TimeoutError: If the total poll time exceeds timeout, raise
        """
        await self.refresh()
        poll_wait = SyncJob._default_poll_wait
        start_time = time.time()
        while not self.has_completed():
            if timeout_secs is not None and (time.time() - start_time) >= timeout_secs:
                raise TimeoutError(f"The wait time has exceeded {timeout_secs} seconds.")

            logger.debug(
                f"Waiting for job {self.id},"
                + f"it is in status '{self.details.status}'"
            )
            if print_progress:
                print(".", end="", flush=True)
            await asyncio.sleep(poll_wait)
            await self.refresh()
            poll_wait = (
                max_poll_wait_secs
                if poll_wait >= max_poll_wait_secs
                else poll_wait * 1.5
            )

    async def get_results(self, timeout_secs: float = DEFAULT_TIMEOUT) -> dict:
        """Get job results by downloading the results blob from the
        storage container linked via the workspace.

        :param timeout_secs: Timeout in seconds, defaults to 300
        :type timeout_secs: int
        :raises RuntimeError: [description]
        :return: [description]
        :rtype: dict
        """
        if self.results is not None:
            return self.results

        if not self.has_completed():
            await self.wait_until_completed(timeout_secs=timeout_secs)

        if not self.details.status == "Succeeded":
            raise RuntimeError(
                f'{"Cannot retrieve results as job execution failed"}'
                + f"(status: {self.details.status}."
                + f"error: {self.details.error_data})"
            )

        payload = await self.download_data(self.details.output_data_uri)
        results = json.loads(payload.decode("utf8"))
        return results



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import abc
import logging
import uuid

from enum import Enum
from urllib.parse import urlparse
from typing import Any, Dict, Optional, TYPE_CHECKING
from azure.storage.blob import BlobClient

from azure.quantum.storage import upload_blob, download_blob, ContainerClient
from azure.quantum._client.models import JobDetails
from azure.quantum.job.workspace_item import WorkspaceItem


if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace


logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 300  # Default timeout for waiting for job to complete

class ContentType(str, Enum):
    json = "application/json"

class BaseJob(WorkspaceItem):
    # Optionally override these to create a Provider-specific Job subclass
    """
    Base job class with methods to create a job from raw blob data,
    upload blob data and download results.
    """

    @staticmethod
    def create_job_id() -> str:
        """Create a unique id for a new job."""
        return str(uuid.uuid1())

    @property
    def details(self) -> JobDetails:
        return self._details

    @details.setter
    def details(self, value: JobDetails):
        self._details = value

    @property
    def container_name(self):
        return f"job-{self.id}"

    @classmethod
    def from_input_data(
        cls,
        workspace: "Workspace",
        name: str,
        target: str,
        input_data: bytes,
        content_type: ContentType = ContentType.json,
        blob_name: str = "inputData",
        encoding: str = "",
        job_id: str = None,
        container_name: str = None,
        provider_id: str = None,
        input_data_format: str = None,
        output_data_format: str = None,
        input_params: Dict[str, Any] = None,
        session_id: Optional[str] = None,
        **kwargs
    ) -> "BaseJob":
        """Create a new Azure Quantum job based on a raw input_data payload.

        :param workspace: Azure Quantum workspace to submit the input_data to
        :type workspace: "Workspace"
        :param name: Name of the job
        :type name: str
        :param target: Azure Quantum target
        :type target: str
        :param input_data: Raw input data to submit
        :type input_data: bytes
        :param blob_name: Input data blob name, defaults to "inputData"
        :type blob_name: str
        :param content_type: Content type, e.g. "application/json"
        :type content_type: ContentType
        :param encoding: input_data encoding, e.g. "gzip", defaults to empty string
        :type encoding: str
        :param job_id: Job ID, defaults to None
        :type job_id: str, optional
        :param container_name: Container name, defaults to None
        :type container_name: str
        :param provider_id: Provider ID, defaults to None
        :type provider_id: str, optional
        :param input_data_format: Input data format, defaults to None
        :type input_data_format: str, optional
        :param output_data_format: Output data format, defaults to None
        :type output_data_format: str, optional
        :param input_params: Input parameters, defaults to None
        :type input_params: Dict[str, Any], optional
        :param input_params: Input params for job
        :type input_params: Dict[str, Any]
        :return: Azure Quantum Job
        :rtype: Job
        """
        # Generate job ID if not specified
        if job_id is None:
            job_id = cls.create_job_id()

        # Create container if it does not yet exist
        container_uri = workspace.get_container_uri(
            job_id=job_id,
            container_name=container_name
        )
        logger.debug(f"Container URI: {container_uri}")

        # Upload data to container
        input_data_uri = cls.upload_input_data(
            container_uri=container_uri,
            input_data=input_data,
            content_type=content_type,
            blob_name=blob_name,
            encoding=encoding,
        )

        # Create and submit job
        return cls.from_storage_uri(
            workspace=workspace,
            job_id=job_id,
            target=target,
            input_data_uri=input_data_uri,
            container_uri=container_uri,
            name=name,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            provider_id=provider_id,
            input_params=input_params,
            session_id=session_id,
            **kwargs
        )

    @classmethod
    def from_storage_uri(
        cls,
        workspace: "Workspace",
        name: str,
        target: str,
        input_data_uri: str,
        provider_id: str,
        input_data_format: str,
        output_data_format: str,
        container_uri: str = None,
        job_id: str = None,
        input_params: Dict[str, Any] = None,
        submit_job: bool = True,
        session_id: Optional[str] = None,
        **kwargs
    ) -> "BaseJob":
        """Create new Job from URI if input data is already uploaded
        to blob storage

        :param workspace: Azure Quantum workspace to submit the blob to
        :type workspace: "Workspace"
        :param name: Job name
        :type name: str
        :param target: Azure Quantum target
        :type target: str
        :param input_data_uri: Input data URI
        :type input_data_uri: str
        :param provider_id: Provider ID
        :type provider_id: str, optional
        :param input_data_format: Input data format
        :type input_data_format: str, optional
        :param output_data_format: Output data format
        :type output_data_format: str, optional
        :param container_uri: Container URI, defaults to None
        :type container_uri: str
        :param job_id: Pre-generated job ID, defaults to None
        :type job_id: str
        :param input_params: Input parameters, defaults to None
        :type input_params: Dict[str, Any], optional
        :param submit_job: If job should be submitted to the service, defaults to True
        :type submit_job: bool
        :return: Job instance
        :rtype: Job
        """
        # Generate job_id, input_params, data formats and provider ID if not specified
        if job_id is None:
            job_id = cls.create_job_id()
        if input_params is None:
            input_params = {}

        # Create container for output data if not specified
        if container_uri is None:
            container_uri = workspace.get_container_uri(job_id=job_id)

        # Create job details and return Job
        details = JobDetails(
            id=job_id,
            name=name,
            container_uri=container_uri,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            input_data_uri=input_data_uri,
            provider_id=provider_id,
            target=target,
            input_params=input_params,
            session_id=session_id,
            **kwargs
        )
        job = cls(workspace, details, **kwargs)

        logger.info(
            f"Submitting problem '{name}'. \
                Using payload from: '{job.details.input_data_uri}'"
        )

        if submit_job:
            logger.debug(f"==> submitting: {job.details}")
            job.submit()

        return job

    @staticmethod
    def upload_input_data(
        container_uri: str,
        input_data: bytes,
        content_type: Optional[ContentType] = ContentType.json,
        blob_name: str = "inputData",
        encoding: str = "",
        return_sas_token: bool = False
    ) -> str:
        """Upload input data file

        :param container_uri: Container URI
        :type container_uri: str
        :param input_data: Input data in binary format
        :type input_data: bytes
        :param content_type: Content type, e.g. "application/json"
        :type content_type: Optional, ContentType
        :param blob_name: Blob name, defaults to "inputData"
        :type blob_name: str, optional
        :param encoding: Encoding, e.g. "gzip", defaults to ""
        :type encoding: str, optional
        :param return_sas_token: Flag to return SAS token as part of URI, defaults to False
        :type return_sas_token: bool, optional
        :return: Uploaded data URI
        :rtype: str
        """
        container_client = ContainerClient.from_container_url(
            container_uri
        )

        uploaded_blob_uri = upload_blob(
            container_client,
            blob_name,
            content_type,
            encoding,
            input_data,
            return_sas_token=return_sas_token
        )
        return uploaded_blob_uri

    def download_data(self, blob_uri: str) -> dict:
        """Download file from blob uri

        :param blob_uri: Blob URI
        :type blob_uri: str
        :return: Payload from blob
        :rtype: dict
        """
        url = urlparse(blob_uri)
        if url.query.find("se=") == -1:
            # blob_uri does not contains SAS token,
            # get sas url from service
            blob_client = BlobClient.from_blob_url(
                blob_uri
            )
            blob_uri = self.workspace._get_linked_storage_sas_uri(
                blob_client.container_name, blob_client.blob_name
            )
            payload = download_blob(blob_uri)
        else:
            # blob_uri contains SAS token, use it
            payload = download_blob(blob_uri)

        return payload

    def upload_attachment(
        self,
        name: str,
        data: bytes,
        container_uri: str = None,
        **kwargs
    ) -> str:
        """Uploads an attachment to the job's container file. Attachment's are identified by name.
        Uploading to an existing attachment overrides its previous content.

        :param name: Attachment name
        :type name: str
        :param data: Attachment data in binary format
        :type input_data: bytes
        :param container_uri: Container URI, defaults to the job's linked container.
        :type container_uri: str, Optional

        :return: Uploaded data URI
        :rtype: str
        """

        # Use Job's default container if not specified
        if container_uri is None:
            container_uri = self.workspace.get_container_uri(job_id=self.id)

        uploaded_blob_uri = self.upload_input_data(
            container_uri = container_uri,
            blob_name = name,
            input_data = data,
            **kwargs
        )
        return uploaded_blob_uri

    def download_attachment(
        self,
        name: str,
        container_uri: str = None
    ):
        """ Downloads an attachment from job's container in Azure Storage. Attachments are blobs of data
            created as part of the Job's execution, or they can be created by uploading directly from Python
            using the upload_attachment method.
            
        :param name: Attachment name
        :type name: str
        :param container_uri: Container URI, defaults to the job's linked container.
        :type container_uri: str, Optional

        :return: Attachment data
        :rtype: bytes
        """

        # Use Job's default container if not specified
        if container_uri is None:
            container_uri = self.workspace.get_container_uri(job_id=self.id)
        
        container_client = ContainerClient.from_container_url(container_uri)
        blob_client = container_client.get_blob_client(name)
        response = blob_client.download_blob().readall()
        return response



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import re
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Type, Union
from ...job import Job
from ...job.base_job import ContentType
from ...workspace import Workspace
from ..params import InputParams, InputParamsItem, AutoValidatingParams, \
    validating_field
from ..target import Target
from . import MicrosoftEstimatorJob
from typing import List


class QubitParams:
    GATE_US_E3 = "qubit_gate_us_e3"
    GATE_US_E4 = "qubit_gate_us_e4"
    GATE_NS_E3 = "qubit_gate_ns_e3"
    GATE_NS_E4 = "qubit_gate_ns_e4"
    MAJ_NS_E4 = "qubit_maj_ns_e4"
    MAJ_NS_E6 = "qubit_maj_ns_e6"


class QECScheme:
    SURFACE_CODE = "surface_code"
    FLOQUET_CODE = "floquet_code"


def _check_error_rate(name, value):
    if value <= 0.0 or value >= 1.0:
        raise ValueError(f"{name} must be between 0 and 1")

def _check_error_rate_or_process_and_readout(name, value):
    if value is None:
        return

    if isinstance(value, float):
        _check_error_rate(name, value)
        return

    if not isinstance(value, MeasurementErrorRate):
        raise ValueError(f"{name} must be either a float or "
                         "MeasurementErrorRate with two fields: 'process' and 'readout'")

def check_time(name, value):
    pat = r"^(\+?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*(s|ms|μs|µs|us|ns)$"
    if re.match(pat, value) is None:
        raise ValueError(f"{name} is not a valid time string; use a "
                         "suffix s, ms, us, or ns")

@dataclass
class MeasurementErrorRate(AutoValidatingParams):
    process: float = field(metadata={"validate": _check_error_rate})
    readout: float = field(metadata={"validate": _check_error_rate})

@dataclass
class MicrosoftEstimatorQubitParams(AutoValidatingParams):
    @staticmethod
    def check_instruction_set(name, value):
        if value not in ["gate-based", "gate_based", "GateBased", "gateBased",
                         "Majorana", "majorana"]:
            raise ValueError(f"{name} must be GateBased or Majorana")

    name: Optional[str] = None
    instruction_set: Optional[str] = validating_field(check_instruction_set)
    one_qubit_measurement_time: Optional[str] = validating_field(check_time)
    two_qubit_joint_measurement_time: Optional[str] = \
        validating_field(check_time)
    one_qubit_gate_time: Optional[str] = validating_field(check_time)
    two_qubit_gate_time: Optional[str] = validating_field(check_time)
    t_gate_time: Optional[str] = validating_field(check_time)
    one_qubit_measurement_error_rate: Union[None, float, MeasurementErrorRate] = \
        validating_field(_check_error_rate_or_process_and_readout)
    two_qubit_joint_measurement_error_rate: Union[None, float, MeasurementErrorRate] = \
        validating_field(_check_error_rate_or_process_and_readout)
    one_qubit_gate_error_rate: Optional[float] = \
        validating_field(_check_error_rate)
    two_qubit_gate_error_rate: Optional[float] = \
        validating_field(_check_error_rate)
    t_gate_error_rate: Optional[float] = validating_field(_check_error_rate)
    idle_error_rate: Optional[float] = validating_field(_check_error_rate)

    _default_models = [QubitParams.GATE_US_E3, QubitParams.GATE_US_E4,
                       QubitParams.GATE_NS_E3, QubitParams.GATE_NS_E4,
                       QubitParams.MAJ_NS_E4, QubitParams.MAJ_NS_E6]
    _gate_based = ["gate-based", "gate_based", "GateBased", "gateBased"]
    _maj_based = ["Majorana", "majorana"]

    def post_validation(self, result):
        # check whether all fields have been specified in case a custom qubit
        # model is specified
        custom = result != {} and \
            (self.name is None or self.name not in self._default_models)

        # no further validation needed for non-custom models
        if not custom:
            return

        # instruction set must be set
        if self.instruction_set is None:
            raise LookupError("instruction_set must be set for custom qubit "
                              "parameters")

        # NOTE at this point, we know that instruction set must have valid
        # value
        if self.one_qubit_measurement_time is None:
            raise LookupError("one_qubit_measurement_time must be set")
        if self.one_qubit_measurement_error_rate is None:
            raise LookupError("one_qubit_measurement_error_rate must be set")

        # this only needs to be checked for gate based qubits
        if self.instruction_set in self._gate_based:
            if self.one_qubit_gate_time is None:
                raise LookupError("one_qubit_gate_time must be set")

    def as_dict(self, validate=True) -> Dict[str, Any]:
        qubit_params = super().as_dict(validate)
        if len(qubit_params) != 0:
            if isinstance(self.one_qubit_measurement_error_rate, MeasurementErrorRate):
                qubit_params["oneQubitMeasurementErrorRate"] = \
                self.one_qubit_measurement_error_rate.as_dict(validate)

            if isinstance(self.two_qubit_joint_measurement_error_rate, MeasurementErrorRate):
                qubit_params["twoQubitJointMeasurementErrorRate"] = \
                self.two_qubit_joint_measurement_error_rate.as_dict(validate)

        return qubit_params


@dataclass
class MicrosoftEstimatorQecScheme(AutoValidatingParams):
    name: Optional[str] = None
    error_correction_threshold: Optional[float] = \
        validating_field(_check_error_rate)
    crossing_prefactor: Optional[float] = None
    logical_cycle_time: Optional[str] = None
    physical_qubits_per_logical_qubit: Optional[str] = None


@dataclass
class ProtocolSpecificDistillationUnitSpecification(AutoValidatingParams):
    num_unit_qubits: Optional[int] = None
    duration_in_qubit_cycle_time: Optional[int] = None

    def post_validation(self, result):
        if self.num_unit_qubits is None:
            raise LookupError("num_unit_qubits must be set")

        if self.duration_in_qubit_cycle_time is None:
            raise LookupError("duration_in_qubit_cycle_time must be set")


@dataclass
class DistillationUnitSpecification(AutoValidatingParams):
    name: Optional[str] = None
    display_name: Optional[str] = None
    num_input_ts: Optional[int] = None
    num_output_ts: Optional[int] = None
    failure_probability_formula:  Optional[str] = None
    output_error_rate_formula:  Optional[str] = None
    physical_qubit_specification: Optional[ProtocolSpecificDistillationUnitSpecification] = None
    logical_qubit_specification: Optional[ProtocolSpecificDistillationUnitSpecification] = None
    logical_qubit_specification_first_round_override: \
        Optional[ProtocolSpecificDistillationUnitSpecification] = None

    def has_custom_specification(self):
        return \
        self.display_name is not None \
        or self.num_input_ts is not None \
        or self.num_output_ts is not None \
        or self.failure_probability_formula is not None \
        or self.output_error_rate_formula is not None \
        or self.physical_qubit_specification is not None \
        or self.logical_qubit_specification is not None \
        or self.logical_qubit_specification_first_round_override is not None

    def has_predefined_name(self):
        return self.name is not None

    def post_validation(self, result):
        if not self.has_custom_specification() and not self.has_predefined_name():
            raise LookupError("name must be set or custom specification must be provided")

        if self.has_custom_specification() and self.has_predefined_name():
            raise LookupError("If predefined name is provided, "
                            "custom specification is not allowed. "
                            "Either remove name or remove all other "
                            "specification of the distillation unit")

        if self.has_predefined_name():
            return # all other validation is on the server side

        if self.num_input_ts is None:
            raise LookupError("num_input_ts must be set")

        if self.num_output_ts is None:
            raise LookupError("num_output_ts must be set")

        if self.failure_probability_formula is None:
            raise LookupError("failure_probability_formula must be set")

        if self.output_error_rate_formula is None:
            raise LookupError("output_error_rate_formula must be set")

        if self.physical_qubit_specification is not None:
            self.physical_qubit_specification.post_validation(result)

        if self.logical_qubit_specification is not None:
            self.logical_qubit_specification.post_validation(result)

        if self.logical_qubit_specification_first_round_override is not None:
            self.logical_qubit_specification_first_round_override.post_validation(result)

    def as_dict(self, validate=True) -> Dict[str, Any]:
        specification_dict = super().as_dict(validate)
        if len(specification_dict) != 0:
            if self.physical_qubit_specification is not None:
                physical_qubit_specification_dict = \
                self.physical_qubit_specification.as_dict(validate)
                if len(physical_qubit_specification_dict) != 0:
                    specification_dict["physicalQubitSpecification"] = \
                        physical_qubit_specification_dict

            if self.logical_qubit_specification is not None:
                logical_qubit_specification_dict = \
                self.logical_qubit_specification.as_dict(validate)
                if len(logical_qubit_specification_dict) != 0:
                    specification_dict["logicalQubitSpecification"] = \
                        logical_qubit_specification_dict

            if self.logical_qubit_specification_first_round_override is not None:
                logical_qubit_specification_first_round_override_dict = \
                self.logical_qubit_specification_first_round_override.as_dict(validate)
                if len(logical_qubit_specification_first_round_override_dict) != 0:
                    specification_dict["logicalQubitSpecificationFirstRoundOverride"] = \
                        logical_qubit_specification_first_round_override_dict

        return specification_dict


@dataclass
class ErrorBudgetPartition(AutoValidatingParams):
    logical: float = 0.001 / 3
    t_states: float = 0.001 / 3
    rotations: float = 0.001 / 3


@dataclass
class MicrosoftEstimatorConstraints(AutoValidatingParams):
    @staticmethod
    def at_least_one(name, value):
        if value < 1:
            raise ValueError(f"{name} must be at least 1")

    logical_depth_factor: Optional[float] = validating_field(at_least_one)
    max_t_factories: Optional[int] = validating_field(at_least_one)
    max_duration: Optional[int] = validating_field(check_time)
    max_physical_qubits: Optional[int] = validating_field(at_least_one)

    def post_validation(self, result):
        if self.max_duration is not None and self.max_physical_qubits is not None:
            raise LookupError("Both duration and number of physical qubits constraints are provided, but only one is allowe at a time.")


@dataclass
class MicrosoftEstimatorProfiling(AutoValidatingParams):
    @staticmethod
    def at_most_30(name, value):
        if value < 0 or value > 30:
            raise ValueError(f"{name} must be nonnegative and at most 30")

    call_stack_depth: Optional[int] = validating_field(at_most_30)
    inline_functions: Optional[bool] = None


class MicrosoftEstimatorInputParamsItem(InputParamsItem):
    """
    Input params for microsoft.estimator target

    :ivar error_budget Total error budget for execution of the algorithm
    """

    def __init__(self):
        super().__init__()

        self.qubit_params: MicrosoftEstimatorQubitParams = \
            MicrosoftEstimatorQubitParams()
        self.qec_scheme: MicrosoftEstimatorQecScheme = \
            MicrosoftEstimatorQecScheme()
        self.distillation_unit_specifications = []  # type: List[DistillationUnitSpecification]
        self.constraints: MicrosoftEstimatorConstraints = \
            MicrosoftEstimatorConstraints()
        self.profiling: MicrosoftEstimatorProfiling = \
            MicrosoftEstimatorProfiling()
        self.error_budget: Optional[Union[float, ErrorBudgetPartition]] = None

    def as_dict(self, validate=True) -> Dict[str, Any]:
        result = super().as_dict(validate)

        qubit_params = self.qubit_params.as_dict(validate)
        if len(qubit_params) != 0:
            result["qubitParams"] = qubit_params

        qec_scheme = self.qec_scheme.as_dict(validate)
        if len(qec_scheme) != 0:
            result["qecScheme"] = qec_scheme

        for specification in self.distillation_unit_specifications:
            specification_dict = specification.as_dict(validate)
            if len(specification_dict) != 0:
                if result.get("distillationUnitSpecifications") is None:
                    result["distillationUnitSpecifications"] = []

                result["distillationUnitSpecifications"].append(specification_dict)

        constraints = self.constraints.as_dict(validate)
        if len(constraints) != 0:
            result["constraints"] = constraints

        profiling = self.profiling.as_dict(validate)
        if len(profiling) != 0:
            result["profiling"] = profiling

        if self.error_budget is not None:
            if isinstance(self.error_budget, float) or \
                    isinstance(self.error_budget, int):
                if validate and \
                        (self.error_budget <= 0 or self.error_budget >= 1):
                    message = "error_budget must be value between 0 and 1"
                    raise ValueError(message)
                result["errorBudget"] = self.error_budget
            elif isinstance(self.error_budget, ErrorBudgetPartition):
                result["errorBudget"] = self.error_budget.as_dict(validate)

        return result


class MicrosoftEstimatorParams(InputParams, MicrosoftEstimatorInputParamsItem):
    def __init__(self, num_items: Optional[int] = None):
        InputParams.__init__(
            self,
            num_items=num_items,
            item_type=MicrosoftEstimatorInputParamsItem)


class MicrosoftEstimator(Target):
    """
    Resource estimator target from the microsoft-qc provider.
    """

    target_names = [
        "microsoft.estimator"
    ]

    def __init__(
        self,
        workspace: "Workspace",
        name: str = "microsoft.estimator",
        **kwargs
    ):
        # There is only a single target name for this target
        assert name == self.target_names[0]

        # make sure to not pass argument twice
        kwargs.pop("provider_id", None)

        super().__init__(
            workspace=workspace,
            name=name,
            input_data_format="qir.v1",
            output_data_format="microsoft.resource-estimates.v1",
            provider_id="microsoft-qc",
            content_type=ContentType.json,
            **kwargs
        )

    def submit(self,
               input_data: Any,
               name: str = "azure-quantum-job",
               input_params: Union[Dict[str, Any], InputParams, None] = None,
               **kwargs) -> Job:
        try:
            from qiskit import QuantumCircuit, transpile
            from qiskit_qir import to_qir_module
            from qiskit_qir.visitor import SUPPORTED_INSTRUCTIONS
            if isinstance(input_data, QuantumCircuit):
                input_data = transpile(input_data,
                                       basis_gates=SUPPORTED_INSTRUCTIONS,
                                       optimization_level=0)
                (module, _) = to_qir_module(input_data, record_output=False)
                input_data = module.bitcode
        finally:
            return super().submit(input_data, name, input_params, **kwargs)

    @classmethod
    def _get_job_class(cls) -> Type[Job]:
        return MicrosoftEstimatorJob

    def _qir_output_data_format(self) -> str:
        """"Fallback output data format in case of QIR job submission."""
        return "microsoft.resource-estimates.v1"

    def make_params(self, num_items: Optional[int] = None):
        return MicrosoftEstimatorParams(num_items=num_items)



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
import os
import json

import logging
import warnings

logger = logging.getLogger(__name__)

from typing import Any, Dict, Tuple, Union, List
from azure.quantum.version import __version__
from azure.quantum.qiskit.job import (
    MICROSOFT_OUTPUT_DATA_FORMAT,
    MICROSOFT_OUTPUT_DATA_FORMAT_V2,
    AzureQuantumJob,
)
from abc import abstractmethod
from azure.quantum.job.session import SessionHost

try:
    from qiskit import QuantumCircuit, transpile
    from qiskit.providers import BackendV1 as Backend
    from qiskit.providers import Options
    from qiskit.providers import Provider
    from qiskit.providers.models import BackendConfiguration
    from qiskit.qobj import Qobj, QasmQobj
    from pyqir import Module
    from qiskit_qir import to_qir_module

except ImportError:
    raise ImportError(
        "Missing optional 'qiskit' dependencies. \
To install run: pip install azure-quantum[qiskit]"
    )


class AzureBackendBase(Backend, SessionHost):
    @abstractmethod
    def __init__(
        self,
        configuration: BackendConfiguration,
        provider: Provider = None,
        **fields
    ):
        super().__init__(configuration, provider, **fields)

    @classmethod
    @abstractmethod
    def _default_options(cls) -> Options:
        pass

    @abstractmethod
    def _azure_config(self) -> Dict[str, str]:
        pass

    def retrieve_job(self, job_id) -> AzureQuantumJob:
        """Returns the Job instance associated with the given id."""
        return self._provider.get_job(job_id)

    def _get_output_data_format(self, options: Dict[str, Any] = {}) -> str:
        config: BackendConfiguration = self.configuration()
        # output data format default depends on the number of experiments. QIR backends
        # that don't define a default in their azure config will use this value
        # Once more than one experiment is supported, we should always use the v2 format
        default_output_data_format = (
            MICROSOFT_OUTPUT_DATA_FORMAT
            if config.max_experiments == 1
            else MICROSOFT_OUTPUT_DATA_FORMAT_V2
        )

        azure_config: Dict[str, Any] = config.azure
        # if the backend defines an output format, use that over the default
        azure_defined_override = azure_config.get(
            "output_data_format", default_output_data_format
        )
        # if the user specifies an output format, use that over the default azure config
        output_data_format = options.pop("output_data_format", azure_defined_override)

        return output_data_format

    def _get_input_params(self, options) -> Dict[str, Any]:
        # Backend options are mapped to input_params.
        input_params: Dict[str, Any] = vars(self.options).copy()

        # The shots/count number can be specified in different ways for different providers,
        # so let's get it first. Values in 'kwargs' take precedence over options, and to keep
        # the convention, 'count' takes precedence over 'shots' afterwards.
        shots_count = (
            options["count"]
            if "count" in options
            else options["shots"]
            if "shots" in options
            else input_params["count"]
            if "count" in input_params
            else input_params["shots"]
            if "shots" in input_params
            else None
        )

        # Let's clear the options of both properties regardless of which one was used to prevent
        # double specification of the value.
        options.pop("shots", None)
        options.pop("count", None)

        if "items" in options:
            input_params["items"] = options.pop("items")

        # Take also into consideration options passed in the options, as the take precedence
        # over default values:
        for opt in options.copy():
            if opt in input_params:
                input_params[opt] = options.pop(opt)

        input_params["count"] = shots_count
        input_params["shots"] = shots_count

        return input_params

    def _run(self, job_name, input_data, input_params, metadata, **options):
        logger.info(f"Submitting new job for backend {self.name()}")

        # The default of these job parameters come from the AzureBackend configuration:
        config = self.configuration()
        blob_name = options.pop("blob_name", config.azure["blob_name"])
        content_type = options.pop("content_type", config.azure["content_type"])
        provider_id = options.pop("provider_id", config.azure["provider_id"])
        input_data_format = options.pop(
            "input_data_format", config.azure["input_data_format"]
        )
        output_data_format = self._get_output_data_format(options)

        # QIR backends will have popped "targetCapability" to configure QIR generation.
        # Anything left here is an invalid parameter with the user attempting to use
        # deprecated parameters.
        targetCapability = input_params.get("targetCapability", None)
        if (
            targetCapability not in [None, "qasm"]
            and input_data_format != "qir.v1"
        ):
            message = "The targetCapability parameter has been deprecated and is only supported for QIR backends."
            message += os.linesep
            message += "To find a QIR capable backend, use the following code:"
            message += os.linesep
            message += (
                f'\tprovider.get_backend("{self.name()}", input_data_format: "qir.v1").'
            )
            raise ValueError(message)

        job = AzureQuantumJob(
            backend=self,
            target=self.name(),
            name=job_name,
            input_data=input_data,
            blob_name=blob_name,
            content_type=content_type,
            provider_id=provider_id,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            input_params=input_params,
            metadata=metadata,
            **options,
        )

        return job

    def _normalize_run_input_params(self, run_input, **options):
        if "circuit" not in options:
            # circuit is not provided, check if there is run_input
            if run_input:
                return run_input
            else:
                raise ValueError("No input provided.")

        if run_input:
            # even though circuit is provided, we still have run_input
            warnings.warn(
                DeprecationWarning(
                    "The circuit parameter has been deprecated and will be ignored."
                )
            )
            return run_input
        else:
            warnings.warn(
                DeprecationWarning(
                    "The circuit parameter has been deprecated. Please use the run_input parameter."
                )
            )

        # we don't have run_input
        # we know we have circuit parameter, but it may be empty
        circuit = options.get("circuit")

        if circuit:
            return circuit
        else:
            raise ValueError("No input provided.")

    def _get_azure_workspace(self) -> "Workspace":
        return self.provider().get_workspace()

    def _get_azure_target_id(self) -> str:
        return self.name()

    def _get_azure_provider_id(self) -> str:
        return self._azure_config()["provider_id"]


class AzureQirBackend(AzureBackendBase):
    @abstractmethod
    def __init__(
        self, configuration: BackendConfiguration, provider: Provider = None, **fields
    ):
        super().__init__(configuration, provider, **fields)

    def _azure_config(self) -> Dict[str, str]:
        return {
            "blob_name": "inputData",
            "content_type": "qir.v1",
            "input_data_format": "qir.v1",
        }

    def run(
        self, run_input: Union[QuantumCircuit, List[QuantumCircuit]] = [], **options
    ) -> AzureQuantumJob:
        """Run on the backend.

        This method returns a
        :class:`~azure.quantum.qiskit.job.AzureQuantumJob` object
        that runs circuits. This is an async call.

        Args:
            run_input (QuantumCircuit or List[QuantumCircuit]): An individual or a
            list of :class:`~qiskit.circuits.QuantumCircuit` to run on the backend.

            options: Any kwarg options to pass to the backend for running the
            config. If a key is also present in the options
            attribute/object then the expectation is that the value
            specified will be used instead of what's set in the options
            object.

        Returns:
            Job: The job object for the run
        """
        run_input = self._normalize_run_input_params(run_input, **options)
        options.pop("run_input", None)
        options.pop("circuit", None)

        circuits = list([])
        if isinstance(run_input, QuantumCircuit):
            circuits = [run_input]
        else:
            circuits = run_input

        max_circuits_per_job = self.configuration().max_experiments
        if len(circuits) > max_circuits_per_job:
            raise NotImplementedError(
                f"This backend only supports running a maximum of {max_circuits_per_job} circuits per job."
            )

        # config normalization
        input_params = self._get_input_params(options)

        shots_count = input_params["count"]

        job_name = ""
        if len(circuits) > 1:
            job_name = f"batch-{len(circuits)}-{shots_count}"
        else:
            job_name = circuits[0].name
        job_name = options.pop("job_name", job_name)

        metadata = options.pop("metadata", self._prepare_job_metadata(circuits))

        input_data = self._translate_input(circuits, input_params)

        job = super()._run(job_name, input_data, input_params, metadata, **options)
        logger.info(
            f"Submitted job with id '{job.id()}' with shot count of {shots_count}:"
        )

        return job

    def _prepare_job_metadata(self, circuits: List[QuantumCircuit]) -> Dict[str, str]:
        """Returns the metadata relative to the given circuits that will be attached to the Job"""
        if len(circuits) == 1:
            circuit: QuantumCircuit = circuits[0]
            return {
                "qiskit": str(True),
                "name": circuit.name,
                "num_qubits": circuit.num_qubits,
                "metadata": json.dumps(circuit.metadata),
            }
        # for batch jobs, we don't want to store the metadata of each circuit
        # we fill out the result header in output processing.
        # These headers don't matter for execution are are only used for
        # result processing.
        return {}

    def _generate_qir(
        self, circuits, targetCapability, **to_qir_kwargs
    ) -> Tuple[Module, List[str]]:

        config = self.configuration()
        # Barriers aren't removed by transpilation and must be explicitly removed in the Qiskit to QIR translation.
        emit_barrier_calls = "barrier" in config.basis_gates
        return to_qir_module(
            circuits,
            targetCapability,
            emit_barrier_calls=emit_barrier_calls,
            **to_qir_kwargs,
        )

    def _get_qir_str(self, circuits, targetCapability, **to_qir_kwargs) -> str:
        module, _ = self._generate_qir(circuits, targetCapability, **to_qir_kwargs)
        return str(module)

    def _translate_input(
        self, circuits: List[QuantumCircuit], input_params: Dict[str, Any]
    ) -> bytes:
        """Translates the input values to the QIR expected by the Backend."""
        logger.info(f"Using QIR as the job's payload format.")
        config = self.configuration()

        # Override QIR translation parameters
        # We will record the output by default, but allow the backend to override this, and allow the user to override the backend.
        to_qir_kwargs = input_params.pop(
            "to_qir_kwargs", config.azure.get("to_qir_kwargs", {"record_output": True})
        )
        targetCapability = input_params.pop(
            "targetCapability",
            self.options.get("targetCapability", "AdaptiveExecution"),
        )

        if logger.isEnabledFor(logging.DEBUG):
            qir = self._get_qir_str(circuits, targetCapability, **to_qir_kwargs)
            logger.debug(f"QIR:\n{qir}")

        # We'll transpile automatically to the supported gates in QIR unless explicitly skipped.
        if not input_params.pop("skipTranspile", False):
            # Set of gates supported by QIR targets.
            circuits = transpile(
                circuits, basis_gates=config.basis_gates, optimization_level=0
            )
            # We'll only log the QIR again if we performed a transpilation.
            if logger.isEnabledFor(logging.DEBUG):
                qir = self._get_qir_str(circuits, targetCapability, **to_qir_kwargs)
                logger.debug(f"QIR (Post-transpilation):\n{qir}")

        (module, entry_points) = self._generate_qir(
            circuits, targetCapability, **to_qir_kwargs
        )

        if not "items" in input_params:
            arguments = input_params.pop("arguments", [])
            input_params["items"] = [
                {"entryPoint": name, "arguments": arguments} for name in entry_points
            ]

        return module.bitcode


class AzureBackend(AzureBackendBase):
    """Base class for interfacing with a backend in Azure Quantum"""

    @abstractmethod
    def __init__(
        self, configuration: BackendConfiguration, provider: Provider = None, **fields
    ):
        super().__init__(configuration, provider, **fields)

    backend_name = None

    def _prepare_job_metadata(self, circuit):
        """Returns the metadata relative to the given circuit that will be attached to the Job"""
        return {
            "qiskit": True,
            "name": circuit.name,
            "num_qubits": circuit.num_qubits,
            "metadata": json.dumps(circuit.metadata),
        }

    @abstractmethod
    def _translate_input(self, circuit):
        pass

    def run(self, run_input=None, **kwargs):
        """Submits the given circuit to run on an Azure Quantum backend."""
        options = kwargs
        circuit = self._normalize_run_input_params(run_input, **options)
        options.pop("run_input", None)
        options.pop("circuit", None)

        # Some Qiskit features require passing lists of circuits, so unpack those here.
        # We currently only support single-experiment jobs.
        if isinstance(circuit, (list, tuple)):
            if len(circuit) > 1:
                raise NotImplementedError("Multi-experiment jobs are not supported!")
            circuit = circuit[0]

        # If the circuit was created using qiskit.assemble,
        # disassemble into QASM here
        if isinstance(circuit, QasmQobj) or isinstance(circuit, Qobj):
            from qiskit.assembler import disassemble

            circuits, run, _ = disassemble(circuit)
            circuit = circuits[0]
            if options.get("shots") is None:
                # Note that the default number of shots for QObj is 1024
                # unless the user specifies the backend.
                options["shots"] = run["shots"]

        # If not provided as options, the values of these parameters
        # are calculated from the circuit itself:
        job_name = options.pop("job_name", circuit.name)
        metadata = options.pop("metadata", self._prepare_job_metadata(circuit))

        input_params = self._get_input_params(options)

        input_data = self._translate_input(circuit)

        job = super()._run(job_name, input_data, input_params, metadata, **options)

        shots_count = input_params["count"]
        logger.info(
            f"Submitted job with id '{job.id()}' for circuit '{circuit.name}' with shot count of {shots_count}:"
        )
        logger.info(input_data)

        return job



# coding=utf-8
# pylint: disable=too-many-lines
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

import sys
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

from .. import _serialization

if sys.version_info >= (3, 9):
    from collections.abc import MutableMapping
else:
    from typing import MutableMapping  # type: ignore  # pylint: disable=ungrouped-imports

if TYPE_CHECKING:
    # pylint: disable=unused-import,ungrouped-imports
    from .. import models as _models
JSON = MutableMapping[str, Any]  # pylint: disable=unsubscriptable-object


class BlobDetails(_serialization.Model):
    """Blob details.

    All required parameters must be populated in order to send to Azure.

    :ivar container_name: The container name. Required.
    :vartype container_name: str
    :ivar blob_name: The blob name.
    :vartype blob_name: str
    """

    _validation = {
        "container_name": {"required": True},
    }

    _attribute_map = {
        "container_name": {"key": "containerName", "type": "str"},
        "blob_name": {"key": "blobName", "type": "str"},
    }

    def __init__(self, *, container_name: str, blob_name: Optional[str] = None, **kwargs: Any) -> None:
        """
        :keyword container_name: The container name. Required.
        :paramtype container_name: str
        :keyword blob_name: The blob name.
        :paramtype blob_name: str
        """
        super().__init__(**kwargs)
        self.container_name = container_name
        self.blob_name = blob_name


class CostEstimate(_serialization.Model):
    """The job cost billed by the provider. The final cost on your bill might be slightly different
    due to added taxes and currency conversion rates.

    :ivar currency_code: The currency code.
    :vartype currency_code: str
    :ivar events: List of usage events.
    :vartype events: list[~azure.quantum._client.models.UsageEvent]
    :ivar estimated_total: The estimated total.
    :vartype estimated_total: float
    """

    _attribute_map = {
        "currency_code": {"key": "currencyCode", "type": "str"},
        "events": {"key": "events", "type": "[UsageEvent]"},
        "estimated_total": {"key": "estimatedTotal", "type": "float"},
    }

    def __init__(
        self,
        *,
        currency_code: Optional[str] = None,
        events: Optional[List["_models.UsageEvent"]] = None,
        estimated_total: Optional[float] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword currency_code: The currency code.
        :paramtype currency_code: str
        :keyword events: List of usage events.
        :paramtype events: list[~azure.quantum._client.models.UsageEvent]
        :keyword estimated_total: The estimated total.
        :paramtype estimated_total: float
        """
        super().__init__(**kwargs)
        self.currency_code = currency_code
        self.events = events
        self.estimated_total = estimated_total


class ErrorData(_serialization.Model):
    """An error response from Azure.

    All required parameters must be populated in order to send to Azure.

    :ivar code: An identifier for the error. Codes are invariant and are intended to be consumed
     programmatically. Required.
    :vartype code: str
    :ivar message: A message describing the error, intended to be suitable for displaying in a user
     interface. Required.
    :vartype message: str
    """

    _validation = {
        "code": {"required": True},
        "message": {"required": True},
    }

    _attribute_map = {
        "code": {"key": "code", "type": "str"},
        "message": {"key": "message", "type": "str"},
    }

    def __init__(self, *, code: str, message: str, **kwargs: Any) -> None:
        """
        :keyword code: An identifier for the error. Codes are invariant and are intended to be consumed
         programmatically. Required.
        :paramtype code: str
        :keyword message: A message describing the error, intended to be suitable for displaying in a
         user interface. Required.
        :paramtype message: str
        """
        super().__init__(**kwargs)
        self.code = code
        self.message = message


class ItemDetails(_serialization.Model):
    """Item details. An item can be a job or a session.

    You probably want to use the sub-classes and not this class directly. Known sub-classes are:
    JobDetails, SessionDetails

    Variables are only populated by the server, and will be ignored when sending a request.

    All required parameters must be populated in order to send to Azure.

    :ivar id: The id of the item. Required.
    :vartype id: str
    :ivar name: The name of the item. It is not required for the name to be unique and it's only
     used for display purposes. Required.
    :vartype name: str
    :ivar provider_id: The unique identifier for the provider. Required.
    :vartype provider_id: str
    :ivar target: The target identifier to run the job. Required.
    :vartype target: str
    :ivar item_type: The type of item. Required. Known values are: "Job" and "Session".
    :vartype item_type: str or ~azure.quantum._client.models.ItemType
    :ivar creation_time: The creation time of the item.
    :vartype creation_time: ~datetime.datetime
    :ivar begin_execution_time: The time when the item began execution.
    :vartype begin_execution_time: ~datetime.datetime
    :ivar end_execution_time: The time when the item finished execution.
    :vartype end_execution_time: ~datetime.datetime
    :ivar cost_estimate: The job cost billed by the provider. The final cost on your bill might be
     slightly different due to added taxes and currency conversion rates.
    :vartype cost_estimate: ~azure.quantum._client.models.CostEstimate
    :ivar error_data: An error response from Azure.
    :vartype error_data: ~azure.quantum._client.models.ErrorData
    """

    _validation = {
        "id": {"required": True},
        "name": {"required": True},
        "provider_id": {"required": True},
        "target": {"required": True},
        "item_type": {"required": True},
        "creation_time": {"readonly": True},
        "begin_execution_time": {"readonly": True},
        "end_execution_time": {"readonly": True},
    }

    _attribute_map = {
        "id": {"key": "id", "type": "str"},
        "name": {"key": "name", "type": "str"},
        "provider_id": {"key": "providerId", "type": "str"},
        "target": {"key": "target", "type": "str"},
        "item_type": {"key": "itemType", "type": "str"},
        "creation_time": {"key": "creationTime", "type": "iso-8601"},
        "begin_execution_time": {"key": "beginExecutionTime", "type": "iso-8601"},
        "end_execution_time": {"key": "endExecutionTime", "type": "iso-8601"},
        "cost_estimate": {"key": "costEstimate", "type": "CostEstimate"},
        "error_data": {"key": "errorData", "type": "ErrorData"},
    }

    _subtype_map = {"item_type": {"Job": "JobDetails", "Session": "SessionDetails"}}

    def __init__(
        self,
        *,
        id: str,  # pylint: disable=redefined-builtin
        name: str,
        provider_id: str,
        target: str,
        cost_estimate: Optional["_models.CostEstimate"] = None,
        error_data: Optional["_models.ErrorData"] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword id: The id of the item. Required.
        :paramtype id: str
        :keyword name: The name of the item. It is not required for the name to be unique and it's only
         used for display purposes. Required.
        :paramtype name: str
        :keyword provider_id: The unique identifier for the provider. Required.
        :paramtype provider_id: str
        :keyword target: The target identifier to run the job. Required.
        :paramtype target: str
        :keyword cost_estimate: The job cost billed by the provider. The final cost on your bill might
         be slightly different due to added taxes and currency conversion rates.
        :paramtype cost_estimate: ~azure.quantum._client.models.CostEstimate
        :keyword error_data: An error response from Azure.
        :paramtype error_data: ~azure.quantum._client.models.ErrorData
        """
        super().__init__(**kwargs)
        self.id = id
        self.name = name
        self.provider_id = provider_id
        self.target = target
        self.item_type: Optional[str] = None
        self.creation_time = None
        self.begin_execution_time = None
        self.end_execution_time = None
        self.cost_estimate = cost_estimate
        self.error_data = error_data


class ItemDetailsList(_serialization.Model):
    """List of item details.

    All required parameters must be populated in order to send to Azure.

    :ivar value: Required.
    :vartype value: list[~azure.quantum._client.models.ItemDetails]
    :ivar next_link: Link to the next page of results.
    :vartype next_link: str
    """

    _validation = {
        "value": {"required": True},
    }

    _attribute_map = {
        "value": {"key": "value", "type": "[ItemDetails]"},
        "next_link": {"key": "nextLink", "type": "str"},
    }

    def __init__(self, *, value: List["_models.ItemDetails"], next_link: Optional[str] = None, **kwargs: Any) -> None:
        """
        :keyword value: Required.
        :paramtype value: list[~azure.quantum._client.models.ItemDetails]
        :keyword next_link: Link to the next page of results.
        :paramtype next_link: str
        """
        super().__init__(**kwargs)
        self.value = value
        self.next_link = next_link


class JobDetails(ItemDetails):  # pylint: disable=too-many-instance-attributes
    """Job details.

    Variables are only populated by the server, and will be ignored when sending a request.

    All required parameters must be populated in order to send to Azure.

    :ivar id: The id of the item. Required.
    :vartype id: str
    :ivar name: The name of the item. It is not required for the name to be unique and it's only
     used for display purposes. Required.
    :vartype name: str
    :ivar provider_id: The unique identifier for the provider. Required.
    :vartype provider_id: str
    :ivar target: The target identifier to run the job. Required.
    :vartype target: str
    :ivar item_type: The type of item. Required. Known values are: "Job" and "Session".
    :vartype item_type: str or ~azure.quantum._client.models.ItemType
    :ivar creation_time: The creation time of the item.
    :vartype creation_time: ~datetime.datetime
    :ivar begin_execution_time: The time when the item began execution.
    :vartype begin_execution_time: ~datetime.datetime
    :ivar end_execution_time: The time when the item finished execution.
    :vartype end_execution_time: ~datetime.datetime
    :ivar cost_estimate: The job cost billed by the provider. The final cost on your bill might be
     slightly different due to added taxes and currency conversion rates.
    :vartype cost_estimate: ~azure.quantum._client.models.CostEstimate
    :ivar error_data: An error response from Azure.
    :vartype error_data: ~azure.quantum._client.models.ErrorData
    :ivar job_type: The type of job. Known values are: "Unknown", "QuantumComputing", and
     "Optimization".
    :vartype job_type: str or ~azure.quantum._client.models.JobType
    :ivar session_id: The ID of the session that the job is part of.
    :vartype session_id: str
    :ivar container_uri: The blob container SAS uri, the container is used to host job data.
     Required.
    :vartype container_uri: str
    :ivar input_data_uri: The input blob SAS uri, if specified, it will override the default input
     blob in the container.
    :vartype input_data_uri: str
    :ivar input_data_format: The format of the input data. Required.
    :vartype input_data_format: str
    :ivar input_params: The input parameters for the job. JSON object used by the target solver. It
     is expected that the size of this object is small and only used to specify parameters for the
     execution target, not the input data.
    :vartype input_params: JSON
    :ivar status: The status of the job. Known values are: "Waiting", "Executing", "Succeeded",
     "Failed", and "Cancelled".
    :vartype status: str or ~azure.quantum._client.models.JobStatus
    :ivar metadata: The job metadata. Metadata provides client the ability to store client-specific
     information.
    :vartype metadata: dict[str, str]
    :ivar output_data_uri: The output blob SAS uri. When a job finishes successfully, results will
     be uploaded to this blob.
    :vartype output_data_uri: str
    :ivar output_data_format: The format of the output data.
    :vartype output_data_format: str
    :ivar cancellation_time: The time when a job was successfully cancelled.
    :vartype cancellation_time: ~datetime.datetime
    :ivar quantum_computing_data: Quantum computing data.
    :vartype quantum_computing_data: ~azure.quantum._client.models.QuantumComputingData
    :ivar tags: List of user-supplied tags associated with the job.
    :vartype tags: list[str]
    """

    _validation = {
        "id": {"required": True},
        "name": {"required": True},
        "provider_id": {"required": True},
        "target": {"required": True},
        "item_type": {"required": True},
        "creation_time": {"readonly": True},
        "begin_execution_time": {"readonly": True},
        "end_execution_time": {"readonly": True},
        "job_type": {"readonly": True},
        "container_uri": {"required": True},
        "input_data_format": {"required": True},
        "status": {"readonly": True},
        "cancellation_time": {"readonly": True},
        "quantum_computing_data": {"readonly": True},
    }

    _attribute_map = {
        "id": {"key": "id", "type": "str"},
        "name": {"key": "name", "type": "str"},
        "provider_id": {"key": "providerId", "type": "str"},
        "target": {"key": "target", "type": "str"},
        "item_type": {"key": "itemType", "type": "str"},
        "creation_time": {"key": "creationTime", "type": "iso-8601"},
        "begin_execution_time": {"key": "beginExecutionTime", "type": "iso-8601"},
        "end_execution_time": {"key": "endExecutionTime", "type": "iso-8601"},
        "cost_estimate": {"key": "costEstimate", "type": "CostEstimate"},
        "error_data": {"key": "errorData", "type": "ErrorData"},
        "job_type": {"key": "jobType", "type": "str"},
        "session_id": {"key": "sessionId", "type": "str"},
        "container_uri": {"key": "containerUri", "type": "str"},
        "input_data_uri": {"key": "inputDataUri", "type": "str"},
        "input_data_format": {"key": "inputDataFormat", "type": "str"},
        "input_params": {"key": "inputParams", "type": "object"},
        "status": {"key": "status", "type": "str"},
        "metadata": {"key": "metadata", "type": "{str}"},
        "output_data_uri": {"key": "outputDataUri", "type": "str"},
        "output_data_format": {"key": "outputDataFormat", "type": "str"},
        "cancellation_time": {"key": "cancellationTime", "type": "iso-8601"},
        "quantum_computing_data": {"key": "quantumComputingData", "type": "QuantumComputingData"},
        "tags": {"key": "tags", "type": "[str]"},
    }

    def __init__(
        self,
        *,
        id: str,  # pylint: disable=redefined-builtin
        name: str,
        provider_id: str,
        target: str,
        container_uri: str,
        input_data_format: str,
        cost_estimate: Optional["_models.CostEstimate"] = None,
        error_data: Optional["_models.ErrorData"] = None,
        session_id: Optional[str] = None,
        input_data_uri: Optional[str] = None,
        input_params: Optional[JSON] = None,
        metadata: Optional[Dict[str, str]] = None,
        output_data_uri: Optional[str] = None,
        output_data_format: Optional[str] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword id: The id of the item. Required.
        :paramtype id: str
        :keyword name: The name of the item. It is not required for the name to be unique and it's only
         used for display purposes. Required.
        :paramtype name: str
        :keyword provider_id: The unique identifier for the provider. Required.
        :paramtype provider_id: str
        :keyword target: The target identifier to run the job. Required.
        :paramtype target: str
        :keyword cost_estimate: The job cost billed by the provider. The final cost on your bill might
         be slightly different due to added taxes and currency conversion rates.
        :paramtype cost_estimate: ~azure.quantum._client.models.CostEstimate
        :keyword error_data: An error response from Azure.
        :paramtype error_data: ~azure.quantum._client.models.ErrorData
        :keyword session_id: The ID of the session that the job is part of.
        :paramtype session_id: str
        :keyword container_uri: The blob container SAS uri, the container is used to host job data.
         Required.
        :paramtype container_uri: str
        :keyword input_data_uri: The input blob SAS uri, if specified, it will override the default
         input blob in the container.
        :paramtype input_data_uri: str
        :keyword input_data_format: The format of the input data. Required.
        :paramtype input_data_format: str
        :keyword input_params: The input parameters for the job. JSON object used by the target solver.
         It is expected that the size of this object is small and only used to specify parameters for
         the execution target, not the input data.
        :paramtype input_params: JSON
        :keyword metadata: The job metadata. Metadata provides client the ability to store
         client-specific information.
        :paramtype metadata: dict[str, str]
        :keyword output_data_uri: The output blob SAS uri. When a job finishes successfully, results
         will be uploaded to this blob.
        :paramtype output_data_uri: str
        :keyword output_data_format: The format of the output data.
        :paramtype output_data_format: str
        :keyword tags: List of user-supplied tags associated with the job.
        :paramtype tags: list[str]
        """
        super().__init__(
            id=id,
            name=name,
            provider_id=provider_id,
            target=target,
            cost_estimate=cost_estimate,
            error_data=error_data,
            **kwargs
        )
        self.item_type: str = "Job"
        self.job_type = None
        self.session_id = session_id
        self.container_uri = container_uri
        self.input_data_uri = input_data_uri
        self.input_data_format = input_data_format
        self.input_params = input_params
        self.status = None
        self.metadata = metadata
        self.output_data_uri = output_data_uri
        self.output_data_format = output_data_format
        self.cancellation_time = None
        self.quantum_computing_data = None
        self.tags = tags


class JobDetailsList(_serialization.Model):
    """List of job details.

    :ivar value:
    :vartype value: list[~azure.quantum._client.models.JobDetails]
    :ivar next_link: Link to the next page of results.
    :vartype next_link: str
    """

    _attribute_map = {
        "value": {"key": "value", "type": "[JobDetails]"},
        "next_link": {"key": "nextLink", "type": "str"},
    }

    def __init__(
        self, *, value: Optional[List["_models.JobDetails"]] = None, next_link: Optional[str] = None, **kwargs: Any
    ) -> None:
        """
        :keyword value:
        :paramtype value: list[~azure.quantum._client.models.JobDetails]
        :keyword next_link: Link to the next page of results.
        :paramtype next_link: str
        """
        super().__init__(**kwargs)
        self.value = value
        self.next_link = next_link


class JsonPatchDocument(_serialization.Model):
    """A JSONPatch document as defined by RFC 6902.

    All required parameters must be populated in order to send to Azure.

    :ivar op: The operation to be performed. Required. Known values are: "add", "remove",
     "replace", "move", "copy", and "test".
    :vartype op: str or ~azure.quantum._client.models.JsonPatchOperation
    :ivar path: A JSON-Pointer. Required.
    :vartype path: str
    :ivar value: A value to be used in the operation on the path.
    :vartype value: JSON
    :ivar from_property: Optional field used in copy and move operations.
    :vartype from_property: str
    """

    _validation = {
        "op": {"required": True},
        "path": {"required": True},
    }

    _attribute_map = {
        "op": {"key": "op", "type": "str"},
        "path": {"key": "path", "type": "str"},
        "value": {"key": "value", "type": "object"},
        "from_property": {"key": "from", "type": "str"},
    }

    def __init__(
        self,
        *,
        op: Union[str, "_models.JsonPatchOperation"],
        path: str,
        value: Optional[JSON] = None,
        from_property: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword op: The operation to be performed. Required. Known values are: "add", "remove",
         "replace", "move", "copy", and "test".
        :paramtype op: str or ~azure.quantum._client.models.JsonPatchOperation
        :keyword path: A JSON-Pointer. Required.
        :paramtype path: str
        :keyword value: A value to be used in the operation on the path.
        :paramtype value: JSON
        :keyword from_property: Optional field used in copy and move operations.
        :paramtype from_property: str
        """
        super().__init__(**kwargs)
        self.op = op
        self.path = path
        self.value = value
        self.from_property = from_property


class ProviderStatus(_serialization.Model):
    """Providers status.

    :ivar id: Provider id.
    :vartype id: str
    :ivar current_availability: Provider availability. Known values are: "Available", "Degraded",
     and "Unavailable".
    :vartype current_availability: str or ~azure.quantum._client.models.ProviderAvailability
    :ivar targets:
    :vartype targets: list[~azure.quantum._client.models.TargetStatus]
    """

    _attribute_map = {
        "id": {"key": "id", "type": "str"},
        "current_availability": {"key": "currentAvailability", "type": "str"},
        "targets": {"key": "targets", "type": "[TargetStatus]"},
    }

    def __init__(
        self,
        *,
        id: Optional[str] = None,  # pylint: disable=redefined-builtin
        current_availability: Optional[Union[str, "_models.ProviderAvailability"]] = None,
        targets: Optional[List["_models.TargetStatus"]] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword id: Provider id.
        :paramtype id: str
        :keyword current_availability: Provider availability. Known values are: "Available",
         "Degraded", and "Unavailable".
        :paramtype current_availability: str or ~azure.quantum._client.models.ProviderAvailability
        :keyword targets:
        :paramtype targets: list[~azure.quantum._client.models.TargetStatus]
        """
        super().__init__(**kwargs)
        self.id = id
        self.current_availability = current_availability
        self.targets = targets


class ProviderStatusList(_serialization.Model):
    """Providers status.

    :ivar value:
    :vartype value: list[~azure.quantum._client.models.ProviderStatus]
    :ivar next_link: Link to the next page of results.
    :vartype next_link: str
    """

    _attribute_map = {
        "value": {"key": "value", "type": "[ProviderStatus]"},
        "next_link": {"key": "nextLink", "type": "str"},
    }

    def __init__(
        self, *, value: Optional[List["_models.ProviderStatus"]] = None, next_link: Optional[str] = None, **kwargs: Any
    ) -> None:
        """
        :keyword value:
        :paramtype value: list[~azure.quantum._client.models.ProviderStatus]
        :keyword next_link: Link to the next page of results.
        :paramtype next_link: str
        """
        super().__init__(**kwargs)
        self.value = value
        self.next_link = next_link


class QuantumComputingData(_serialization.Model):
    """Quantum computing data.

    :ivar count: The number of quantum computing items in the job.
    :vartype count: int
    """

    _attribute_map = {
        "count": {"key": "count", "type": "int"},
    }

    def __init__(self, *, count: Optional[int] = None, **kwargs: Any) -> None:
        """
        :keyword count: The number of quantum computing items in the job.
        :paramtype count: int
        """
        super().__init__(**kwargs)
        self.count = count


class Quota(_serialization.Model):
    """Quota information.

    :ivar dimension: The name of the dimension associated with the quota.
    :vartype dimension: str
    :ivar scope: The scope at which the quota is applied. Known values are: "Workspace" and
     "Subscription".
    :vartype scope: str or ~azure.quantum._client.models.DimensionScope
    :ivar provider_id: The unique identifier for the provider.
    :vartype provider_id: str
    :ivar utilization: The amount of the usage that has been applied for the current period.
    :vartype utilization: float
    :ivar holds: The amount of the usage that has been reserved but not applied for the current
     period.
    :vartype holds: float
    :ivar limit: The maximum amount of usage allowed for the current period.
    :vartype limit: float
    :ivar period: The time period in which the quota's underlying meter is accumulated. Based on
     calendar year. 'None' is used for concurrent quotas. Known values are: "None" and "Monthly".
    :vartype period: str or ~azure.quantum._client.models.MeterPeriod
    """

    _attribute_map = {
        "dimension": {"key": "dimension", "type": "str"},
        "scope": {"key": "scope", "type": "str"},
        "provider_id": {"key": "providerId", "type": "str"},
        "utilization": {"key": "utilization", "type": "float"},
        "holds": {"key": "holds", "type": "float"},
        "limit": {"key": "limit", "type": "float"},
        "period": {"key": "period", "type": "str"},
    }

    def __init__(
        self,
        *,
        dimension: Optional[str] = None,
        scope: Optional[Union[str, "_models.DimensionScope"]] = None,
        provider_id: Optional[str] = None,
        utilization: Optional[float] = None,
        holds: Optional[float] = None,
        limit: Optional[float] = None,
        period: Optional[Union[str, "_models.MeterPeriod"]] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword dimension: The name of the dimension associated with the quota.
        :paramtype dimension: str
        :keyword scope: The scope at which the quota is applied. Known values are: "Workspace" and
         "Subscription".
        :paramtype scope: str or ~azure.quantum._client.models.DimensionScope
        :keyword provider_id: The unique identifier for the provider.
        :paramtype provider_id: str
        :keyword utilization: The amount of the usage that has been applied for the current period.
        :paramtype utilization: float
        :keyword holds: The amount of the usage that has been reserved but not applied for the current
         period.
        :paramtype holds: float
        :keyword limit: The maximum amount of usage allowed for the current period.
        :paramtype limit: float
        :keyword period: The time period in which the quota's underlying meter is accumulated. Based on
         calendar year. 'None' is used for concurrent quotas. Known values are: "None" and "Monthly".
        :paramtype period: str or ~azure.quantum._client.models.MeterPeriod
        """
        super().__init__(**kwargs)
        self.dimension = dimension
        self.scope = scope
        self.provider_id = provider_id
        self.utilization = utilization
        self.holds = holds
        self.limit = limit
        self.period = period


class QuotaList(_serialization.Model):
    """List of quotas.

    :ivar value:
    :vartype value: list[~azure.quantum._client.models.Quota]
    :ivar next_link: Link to the next page of results.
    :vartype next_link: str
    """

    _attribute_map = {
        "value": {"key": "value", "type": "[Quota]"},
        "next_link": {"key": "nextLink", "type": "str"},
    }

    def __init__(
        self, *, value: Optional[List["_models.Quota"]] = None, next_link: Optional[str] = None, **kwargs: Any
    ) -> None:
        """
        :keyword value:
        :paramtype value: list[~azure.quantum._client.models.Quota]
        :keyword next_link: Link to the next page of results.
        :paramtype next_link: str
        """
        super().__init__(**kwargs)
        self.value = value
        self.next_link = next_link


class RestError(_serialization.Model):
    """Error information returned by the API.

    All required parameters must be populated in order to send to Azure.

    :ivar error: An error response from Azure. Required.
    :vartype error: ~azure.quantum._client.models.ErrorData
    """

    _validation = {
        "error": {"required": True},
    }

    _attribute_map = {
        "error": {"key": "error", "type": "ErrorData"},
    }

    def __init__(self, *, error: "_models.ErrorData", **kwargs: Any) -> None:
        """
        :keyword error: An error response from Azure. Required.
        :paramtype error: ~azure.quantum._client.models.ErrorData
        """
        super().__init__(**kwargs)
        self.error = error


class SasUriResponse(_serialization.Model):
    """Get SAS URL operation response.

    :ivar sas_uri: A URL with a SAS token to upload a blob for execution in the given workspace.
    :vartype sas_uri: str
    """

    _attribute_map = {
        "sas_uri": {"key": "sasUri", "type": "str"},
    }

    def __init__(self, *, sas_uri: Optional[str] = None, **kwargs: Any) -> None:
        """
        :keyword sas_uri: A URL with a SAS token to upload a blob for execution in the given workspace.
        :paramtype sas_uri: str
        """
        super().__init__(**kwargs)
        self.sas_uri = sas_uri


class SessionDetails(ItemDetails):  # pylint: disable=too-many-instance-attributes
    """Session details.

    Variables are only populated by the server, and will be ignored when sending a request.

    All required parameters must be populated in order to send to Azure.

    :ivar id: The id of the item. Required.
    :vartype id: str
    :ivar name: The name of the item. It is not required for the name to be unique and it's only
     used for display purposes. Required.
    :vartype name: str
    :ivar provider_id: The unique identifier for the provider. Required.
    :vartype provider_id: str
    :ivar target: The target identifier to run the job. Required.
    :vartype target: str
    :ivar item_type: The type of item. Required. Known values are: "Job" and "Session".
    :vartype item_type: str or ~azure.quantum._client.models.ItemType
    :ivar creation_time: The creation time of the item.
    :vartype creation_time: ~datetime.datetime
    :ivar begin_execution_time: The time when the item began execution.
    :vartype begin_execution_time: ~datetime.datetime
    :ivar end_execution_time: The time when the item finished execution.
    :vartype end_execution_time: ~datetime.datetime
    :ivar cost_estimate: The job cost billed by the provider. The final cost on your bill might be
     slightly different due to added taxes and currency conversion rates.
    :vartype cost_estimate: ~azure.quantum._client.models.CostEstimate
    :ivar error_data: An error response from Azure.
    :vartype error_data: ~azure.quantum._client.models.ErrorData
    :ivar job_failure_policy: Policy controlling the behavior of the Session when a job in the
     session fails. Known values are: "Abort" and "Continue".
    :vartype job_failure_policy: str or ~azure.quantum._client.models.SessionJobFailurePolicy
    :ivar status: The status of the session. Known values are: "Waiting", "Executing", "Succeeded",
     "Failed", "Failure(s)", and "TimedOut".
    :vartype status: str or ~azure.quantum._client.models.SessionStatus
    """

    _validation = {
        "id": {"required": True},
        "name": {"required": True},
        "provider_id": {"required": True},
        "target": {"required": True},
        "item_type": {"required": True},
        "creation_time": {"readonly": True},
        "begin_execution_time": {"readonly": True},
        "end_execution_time": {"readonly": True},
        "status": {"readonly": True},
    }

    _attribute_map = {
        "id": {"key": "id", "type": "str"},
        "name": {"key": "name", "type": "str"},
        "provider_id": {"key": "providerId", "type": "str"},
        "target": {"key": "target", "type": "str"},
        "item_type": {"key": "itemType", "type": "str"},
        "creation_time": {"key": "creationTime", "type": "iso-8601"},
        "begin_execution_time": {"key": "beginExecutionTime", "type": "iso-8601"},
        "end_execution_time": {"key": "endExecutionTime", "type": "iso-8601"},
        "cost_estimate": {"key": "costEstimate", "type": "CostEstimate"},
        "error_data": {"key": "errorData", "type": "ErrorData"},
        "job_failure_policy": {"key": "jobFailurePolicy", "type": "str"},
        "status": {"key": "status", "type": "str"},
    }

    def __init__(
        self,
        *,
        id: str,  # pylint: disable=redefined-builtin
        name: str,
        provider_id: str,
        target: str,
        cost_estimate: Optional["_models.CostEstimate"] = None,
        error_data: Optional["_models.ErrorData"] = None,
        job_failure_policy: Union[str, "_models.SessionJobFailurePolicy"] = "Abort",
        **kwargs: Any
    ) -> None:
        """
        :keyword id: The id of the item. Required.
        :paramtype id: str
        :keyword name: The name of the item. It is not required for the name to be unique and it's only
         used for display purposes. Required.
        :paramtype name: str
        :keyword provider_id: The unique identifier for the provider. Required.
        :paramtype provider_id: str
        :keyword target: The target identifier to run the job. Required.
        :paramtype target: str
        :keyword cost_estimate: The job cost billed by the provider. The final cost on your bill might
         be slightly different due to added taxes and currency conversion rates.
        :paramtype cost_estimate: ~azure.quantum._client.models.CostEstimate
        :keyword error_data: An error response from Azure.
        :paramtype error_data: ~azure.quantum._client.models.ErrorData
        :keyword job_failure_policy: Policy controlling the behavior of the Session when a job in the
         session fails. Known values are: "Abort" and "Continue".
        :paramtype job_failure_policy: str or ~azure.quantum._client.models.SessionJobFailurePolicy
        """
        super().__init__(
            id=id,
            name=name,
            provider_id=provider_id,
            target=target,
            cost_estimate=cost_estimate,
            error_data=error_data,
            **kwargs
        )
        self.item_type: str = "Session"
        self.job_failure_policy = job_failure_policy
        self.status = None


class SessionDetailsList(_serialization.Model):
    """List of session details.

    :ivar value:
    :vartype value: list[~azure.quantum._client.models.SessionDetails]
    :ivar next_link: Link to the next page of results.
    :vartype next_link: str
    """

    _attribute_map = {
        "value": {"key": "value", "type": "[SessionDetails]"},
        "next_link": {"key": "nextLink", "type": "str"},
    }

    def __init__(
        self, *, value: Optional[List["_models.SessionDetails"]] = None, next_link: Optional[str] = None, **kwargs: Any
    ) -> None:
        """
        :keyword value:
        :paramtype value: list[~azure.quantum._client.models.SessionDetails]
        :keyword next_link: Link to the next page of results.
        :paramtype next_link: str
        """
        super().__init__(**kwargs)
        self.value = value
        self.next_link = next_link


class TargetStatus(_serialization.Model):
    """Target status.

    :ivar id: Target id.
    :vartype id: str
    :ivar current_availability: Target availability. Known values are: "Available", "Degraded", and
     "Unavailable".
    :vartype current_availability: str or ~azure.quantum._client.models.TargetAvailability
    :ivar average_queue_time: Average queue time in seconds.
    :vartype average_queue_time: int
    :ivar status_page: A page with detailed status of the provider.
    :vartype status_page: str
    """

    _attribute_map = {
        "id": {"key": "id", "type": "str"},
        "current_availability": {"key": "currentAvailability", "type": "str"},
        "average_queue_time": {"key": "averageQueueTime", "type": "int"},
        "status_page": {"key": "statusPage", "type": "str"},
    }

    def __init__(
        self,
        *,
        id: Optional[str] = None,  # pylint: disable=redefined-builtin
        current_availability: Optional[Union[str, "_models.TargetAvailability"]] = None,
        average_queue_time: Optional[int] = None,
        status_page: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword id: Target id.
        :paramtype id: str
        :keyword current_availability: Target availability. Known values are: "Available", "Degraded",
         and "Unavailable".
        :paramtype current_availability: str or ~azure.quantum._client.models.TargetAvailability
        :keyword average_queue_time: Average queue time in seconds.
        :paramtype average_queue_time: int
        :keyword status_page: A page with detailed status of the provider.
        :paramtype status_page: str
        """
        super().__init__(**kwargs)
        self.id = id
        self.current_availability = current_availability
        self.average_queue_time = average_queue_time
        self.status_page = status_page


class UsageEvent(_serialization.Model):
    """Usage event details.

    :ivar dimension_id: The dimension id.
    :vartype dimension_id: str
    :ivar dimension_name: The dimension name.
    :vartype dimension_name: str
    :ivar measure_unit: The unit of measure.
    :vartype measure_unit: str
    :ivar amount_billed: The amount billed.
    :vartype amount_billed: float
    :ivar amount_consumed: The amount consumed.
    :vartype amount_consumed: float
    :ivar unit_price: The unit price.
    :vartype unit_price: float
    """

    _attribute_map = {
        "dimension_id": {"key": "dimensionId", "type": "str"},
        "dimension_name": {"key": "dimensionName", "type": "str"},
        "measure_unit": {"key": "measureUnit", "type": "str"},
        "amount_billed": {"key": "amountBilled", "type": "float"},
        "amount_consumed": {"key": "amountConsumed", "type": "float"},
        "unit_price": {"key": "unitPrice", "type": "float"},
    }

    def __init__(
        self,
        *,
        dimension_id: Optional[str] = None,
        dimension_name: Optional[str] = None,
        measure_unit: Optional[str] = None,
        amount_billed: Optional[float] = None,
        amount_consumed: Optional[float] = None,
        unit_price: Optional[float] = None,
        **kwargs: Any
    ) -> None:
        """
        :keyword dimension_id: The dimension id.
        :paramtype dimension_id: str
        :keyword dimension_name: The dimension name.
        :paramtype dimension_name: str
        :keyword measure_unit: The unit of measure.
        :paramtype measure_unit: str
        :keyword amount_billed: The amount billed.
        :paramtype amount_billed: float
        :keyword amount_consumed: The amount consumed.
        :paramtype amount_consumed: float
        :keyword unit_price: The unit price.
        :paramtype unit_price: float
        """
        super().__init__(**kwargs)
        self.dimension_id = dimension_id
        self.dimension_name = dimension_name
        self.measure_unit = measure_unit
        self.amount_billed = amount_billed
        self.amount_consumed = amount_consumed
        self.unit_price = unit_price



"""Defines targets and helper functions for the Pasqal provider"""

##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

__all__ = [
    "InputParams",
    "Result",
    "Pasqal",
    "PasqalTarget",
]

from .result import Result
from .target import InputParams, Pasqal, PasqalTarget



# coding=utf-8
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import warnings
warnings.warn("The azure.quantum.optimization.toshiba namespace will be deprecated. \
Please use azure.quantum.target.toshiba instead.")

from .solvers import *



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import re
import abc

from typing import Optional
from datetime import date, datetime, timezone

from azure.quantum._client.models import JobStatus


class FilteredJob(abc.ABC):
    """
    Mixin for adding methods to filter jobs
    """
    def matches_filter(
        self, 
        name_match: str = None, 
        status:  Optional[JobStatus] = None,
        created_after: Optional[datetime] = None
    ) -> bool:
        """Checks if job (self) matches the given properties if any.
            :param name_match: regex expression for job name matching
            :param status: filter by job status
            :param created_after: filter jobs after time of job creation
        """
        if name_match is not None and re.search(name_match, self.details.name) is None:
           return False
        
        if status is not None and self.details.status != status.value:
            return False
        
        if created_after is not None:
            # if supplied date is date we must convert to datetime first
            if isinstance(created_after, date):
                created_after = datetime(created_after.year, created_after.month, created_after.day)
            
            # if supplied date is naive, assume local and convert to timezone aware object
            if created_after.tzinfo is None:
                created_after = created_after.astimezone()
            
            if self.details.creation_time.replace(tzinfo=timezone.utc) < created_after:
                return False

        return True



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging
from azure.quantum.version import __version__

from .job.job import *
from .workspace import *

from azure.quantum._client.models._enums import JobStatus

logger = logging.getLogger(__name__)
logger.info(f"version: {__version__}")



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from __future__ import annotations
import logging
import io
import gzip
import json
import numpy
import os
import tarfile


from typing import List, Tuple, Union, Dict, Optional, TYPE_CHECKING
from enum import Enum
from azure.quantum.optimization import TermBase, Term
from azure.quantum.storage import (
    ContainerClient,
    download_blob,
    BlobClient,
    download_blob_metadata,
    download_blob_properties
)
from azure.quantum.job.base_job import ContentType
from azure.quantum.job.job import Job
from azure.quantum.target.target import Target

logger = logging.getLogger(__name__)

__all__ = ["Problem", "ProblemType"]

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace

class ProblemType(str, Enum):
    pubo = 0
    ising = 1


class Problem:
    """Problem to submit to the service.

    :param name: Problem name
    :type name: str
        When passing None, it will default to "Optimization problem"
    :param terms: Problem terms, depending on solver.
        Defaults to None
    :type terms: Optional[List[TermBase]], optional
    :param init_config: Optional configuration details, depending on solver.
        Defaults to None
    :type init_config: Optional[Dict[str,int]], optional
    :param problem_type: Problem type (ProblemType.pubo or
        ProblemType.ising), defaults to ProblemType.ising
    :type problem_type: ProblemType, optional
    :param content_type: Content type, eg: application/json. Default is application/json
    :type content_type: ContentType, optional
    """

    def __init__(
        self,
        name: str,
        terms: Optional[List[TermBase]] = None,
        init_config: Optional[Dict[str, int]] = None,
        problem_type: ProblemType = ProblemType.ising,
        content_type: Optional[ContentType] = ContentType.json 
    ):
        self.name = name or "Optimization problem"
        self.problem_type = problem_type
        self.init_config = init_config
        self.uploaded_blob_uri = None
        self.uploaded_blob_params = None
        self.content_type = content_type

        # each type of term has its own section for quicker serialization
        self.terms = []

        # set the terms
        if terms:
            for term in terms:
                self.terms.append(term)

    """
    Constant thresholds used to determine if a problem is "large".
    """
    NUM_VARIABLES_LARGE = 2500
    NUM_TERMS_LARGE = 1e6


    def serialize(self) -> Union[str, list]:
        """Wrapper function for serializing to json
        """ 
        return self.to_json()

    def to_json(self) -> str:
        """Serializes the problem to a JSON string"""
        result = {
            "metadata": {
                "name": self.name,
            },
            "cost_function": {
                "version": "1.1" if self.init_config else "1.0",
                "type": self.problem_type.name,
                "terms": [term.to_dict() for term in self.terms]
            }
        }

        if self.init_config:
            result["cost_function"]["initial_configuration"] = self.init_config

        return json.dumps(result)

    
    @classmethod
    def from_json(
            cls, 
            input_problem: str, 
            name: Optional[str] = None
        ) -> Problem:
        """Deserializes the problem from a
        json serialized with Problem.serialize()

        :param input_problem:
            the json string to be deserialized to a `Problem` instance
        :type problem_msgs: str
        :param
        :param name: 
            The name of the problem is optional, since it will try 
            to read the serialized name from the json payload.
            If this parameter is not empty, it will use it as the
            problem name ignoring the serialized value.
        :type name: Optional[str]
        """
        result = json.loads(input_problem)

        if name is None:
            metadata = result.get("metadata")
            if metadata is not None:
                name = metadata.get("name")

        terms = [Term.from_dict(t) for t in result["cost_function"]["terms"]] if "terms" in result["cost_function"] else []

        problem = cls(
            name=name,
            terms=terms,
            problem_type=ProblemType[result["cost_function"]["type"]],
        )

        if "initial_configuration" in result["cost_function"]:
            problem.init_config = result["cost_function"]["initial_configuration"]

        return problem

    @classmethod
    def deserialize(
        cls, 
        input_problem: Union[str, list],
        name: Optional[str] = None, 
        content_type: Optional[ContentType] = None) -> Problem:
        """Deserializes the problem from a
        JSON string serialized with Problem.serialize()
        Also used to deserialize the messages downloaded from the blob

        :param input_problem:
            The json string
        :type input_problem: Union[str,list]
        :param
        :param name: 
            The name of the problem is optional, since it will try 
            to read the serialized name from the json payload.
            If this parameter is not empty, it will use it as the
            problem name ignoring the serialized value.
        :type name: Optional[str]
        :param content_type: The content type of the input problem data
        :type: Optional, ContentType
        """
        return cls.from_json(input_problem, name)

    def add_term(self, c: Union[int, float], indices: List[int]):
        """Adds a single monomial term to the `Problem` representation

        :param c: The cost or weight of this term
        :type c: int, float
        :param indices: The variable indices that are in this term
        :type indices: List[int]
        """
        self.terms.append(Term(indices=indices, c=c))
        self.uploaded_blob_uri = None

    def add_terms(
        self, 
        terms: List[Term]
    ):
        """Adds a list of monomial terms to the `Problem` representation

        :param terms: The list of terms to add to the problem
        """
        self.terms += terms
        self.uploaded_blob_uri = None
    
            
    def to_blob(self) -> bytes:
        """Convert problem data to a binary blob.

        :return: Blob data
        :rtype: bytes
        """
        input_problem = self.serialize()
        debug_input_string = input_problem if type(input_problem) is str else b''.join( input_problem).decode('latin-1')
        logger.debug("Input Problem: " + debug_input_string)
        data = io.BytesIO()
        with gzip.GzipFile(fileobj=data, mode="w") as fo:
            fo.write(input_problem.encode())

        return data.getvalue()
    
    def _blob_name(self):
        import uuid
        return "{}-{}".format(self.name, uuid.uuid1())

    def upload(
        self,
        workspace: "Workspace",
        container_name: str = "optimization-problems",
        blob_name: str = "inputData",
        container_uri: str = None,
    ):
        """Uploads an optimization problem instance to
        the cloud storage linked with the Workspace.

        :param workspace: interaction terms of the problem.
        :type workspace: Workspace
        :param container_name: Container name, defaults to "optimization-problems"
        :type container_name: str, optional
        :param blob_name: Blob name, defaults to None
        :type blob_name: str, optional
        :param container_uri: Optional container URI
        :type container_uri: str
        :return: uri of the uploaded problem
        :rtype: str
        """
        blob_params = [workspace, container_name, blob_name]
        if self.uploaded_blob_uri and self.uploaded_blob_params == blob_params:
            return self.uploaded_blob_uri

        if blob_name is None:
            blob_name = self._blob_name()

        encoding = "gzip"
        content_type = self.content_type

        blob = self.to_blob()
        if container_uri is None:
            container_uri = workspace.get_container_uri(
                container_name=container_name
            )
        input_data_uri = Job.upload_input_data(
            input_data=blob,
            blob_name=blob_name,
            container_uri=container_uri,
            encoding=encoding,
            content_type= content_type
        )
        self.uploaded_blob_params = blob_params
        self.uploaded_blob_uri = input_data_uri
        return input_data_uri

    def set_fixed_variables(
        self, fixed_variables: Union[Dict[int, int], Dict[str, int]]
    ) -> Problem:
        """Transforms the current problem with a set of fixed
        variables and returns the new modified problem.
        The original Problem instance is untouched.

        :param fixed_variables:
            The dictionary of variable ids and their fixed state
        """
        if len(fixed_variables) == 0:
            raise RuntimeError(
                "Error: fixed_variables is empty - \
                please specify at least one fixed variable"
            )

        fixed_transformed = {
            int(k): fixed_variables[k] for k in fixed_variables
        }  # if ids are given in string form, convert them to int
        new_terms = []

        constant = 0
        for term in self.terms:
            reduced_term = term.reduce_by_variable_state(fixed_transformed)
            if reduced_term:
                if not isinstance(reduced_term, Term) or len(reduced_term.ids) > 0:
                    new_terms.append(reduced_term)
                else:
                    # reduced to a constant term
                    constant += reduced_term.c

        if constant:
            new_terms.append(Term(c=constant, indices=[]))

        new_init_config = None
        if self.init_config:
            new_init_config = {
                k: self.init_config[k]
                for k in self.init_config
                if int(k) not in fixed_transformed
            }

        return Problem(
            self.name,
            terms=new_terms,
            init_config=new_init_config,
            problem_type=self.problem_type,
        )

    def _evaluate(self, configuration, term_list):
        total = 0
        if term_list:
            for term in term_list:
                total += term.evaluate(configuration)
        return total

    def evaluate(self, configuration: Union[Dict[int, int], Dict[str, int]]) -> float:
        """Given a configuration/variable assignment,
        return the cost function value of this problem.

        :param configuration: The dictionary of
         variable ids to their assigned value
        """
        configuration_transformed = {
            int(k): configuration[k] for k in configuration
        }  # if ids are given in string form, convert them to int

        total_cost = 0
        for terms in self.terms:
            total_cost += self._evaluate(configuration_transformed, terms)

        return total_cost

    def is_large(self) -> bool:
        """Determines if the current problem is large.
        "large" is an arbitrary threshold and can be easily changed.
        Based on usage data, we have defined a
        large problem to be NUM_VARIABLES_LARGE+
        variables AND NUM_TERMS_LARGE+ terms.
        """

        set_vars = set()
        total_term_count = 0
        for term in self.terms:
            if isinstance(term, Term):
                set_vars.update(term.ids)
                total_term_count += 1
        
        return (
            len(set_vars) >= Problem.NUM_VARIABLES_LARGE
            and total_term_count >= Problem.NUM_TERMS_LARGE
        )

    def download(self, workspace: "Workspace"):
        """Downloads the uploaded problem as an instance of `Problem`"""
        if not self.uploaded_blob_uri:
            raise Exception("Problem may not be downloaded before it is uploaded")
        blob_client = BlobClient.from_blob_url(self.uploaded_blob_uri)
        container_client = ContainerClient.from_container_url(
            workspace._get_linked_storage_sas_uri(blob_client.container_name)
        )
        blob_name = blob_client.blob_name
        blob = container_client.get_blob_client(blob_name)
        contents = download_blob(blob.url)
        blob_properties = download_blob_properties(blob.url)        
        content_type = blob_properties.content_type
        return Problem.deserialize(contents, self.name, content_type)

    def get_terms(self, id: int) -> List[TermBase]:
        """Given an index the function will return
        a list of terms with that index
        """
        terms = []
        if self.terms != []:
            for term in self.terms:
                if isinstance(term, Term):
                    if id in term.ids:
                        terms.append(term)
            return terms
        else:
            raise Exception(
                "There are currently no terms in this problem. \
                Please download the problem on the client or add terms to the \
                    problem to perform this operation"
            )


##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from azure.quantum.qiskit.backends.ionq import (
    IonQBackend,
    IonQQPUBackend,
    IonQAriaBackend,
    IonQSimulatorBackend,
    IonQQPUQirBackend,
    IonQAriaQirBackend,
    IonQForteBackend,
    IonQForteQirBackend,
    IonQSimulatorQirBackend,
)

from azure.quantum.qiskit.backends.quantinuum import (
    QuantinuumBackend,
    QuantinuumQPUBackend,
    QuantinuumSyntaxCheckerBackend,
    QuantinuumEmulatorBackend,
    QuantinuumQPUQirBackend,
    QuantinuumSyntaxCheckerQirBackend,
    QuantinuumEmulatorQirBackend,
)

from azure.quantum.qiskit.backends.rigetti import (
    RigettiBackend,
    RigettiQPUBackend,
    RigettiSimulatorBackend,
)

from azure.quantum.qiskit.backends.qci import (
    QCIBackend,
    QCISimulatorBackend,
    QCIQPUBackend,
)

from azure.quantum.qiskit.backends.microsoft import (
    MicrosoftBackend,
    MicrosoftResourceEstimationBackend,
)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from ._models import BlobDetails
from ._models import CostEstimate
from ._models import ErrorData
from ._models import ItemDetails
from ._models import JobDetails
from ._models import JsonPatchDocument
from ._models import ProviderStatus
from ._models import QuantumComputingData
from ._models import Quota
from ._models import RestError
from ._models import SasUriResponse
from ._models import SessionDetails
from ._models import TargetStatus
from ._models import UsageEvent

from ._enums import DimensionScope
from ._enums import ItemType
from ._enums import JobStatus
from ._enums import JobType
from ._enums import JsonPatchOperation
from ._enums import MeterPeriod
from ._enums import ProviderAvailability
from ._enums import SessionJobFailurePolicy
from ._enums import SessionStatus
from ._enums import TargetAvailability
from ._patch import __all__ as _patch_all
from ._patch import *  # pylint: disable=unused-wildcard-import
from ._patch import patch_sdk as _patch_sdk

__all__ = [
    "BlobDetails",
    "CostEstimate",
    "ErrorData",
    "ItemDetails",
    "JobDetails",
    "JsonPatchDocument",
    "ProviderStatus",
    "QuantumComputingData",
    "Quota",
    "RestError",
    "SasUriResponse",
    "SessionDetails",
    "TargetStatus",
    "UsageEvent",
    "DimensionScope",
    "ItemType",
    "JobStatus",
    "JobType",
    "JsonPatchOperation",
    "MeterPeriod",
    "ProviderAvailability",
    "SessionJobFailurePolicy",
    "SessionStatus",
    "TargetAvailability",
]
__all__.extend([p for p in _patch_all if p not in __all__])
_patch_sdk()



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from typing import TYPE_CHECKING, Union

from azure.quantum._client.models import SessionDetails, JobDetails
from azure.quantum.job.job import Job
from azure.quantum.job.session import Session

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace

__all__ = ["WorkspaceItemFactory"]

class WorkspaceItemFactory():
    """

    :param workspace: Workspace instance to submit job to
    :type workspace: Workspace
    :param item_details: Item details model,
            contains item ID, name and other details
    :type item_details: ItemDetails
    """

    @staticmethod
    def __new__(workspace:"Workspace",
                item_details:Union[SessionDetails, JobDetails]
               ) -> Union[Session, Job]:
        if isinstance(item_details, JobDetails):
            return Job(workspace, job_details=item_details)
        elif isinstance(item_details, SessionDetails):
            return Session(workspace, details=item_details)
        else:
            raise TypeError("item_details must be of type `SessionDetails` or `JobDetails`.")



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import warnings
import logging

from typing import TYPE_CHECKING, Union, Any, Optional
from enum import Enum
from azure.quantum.job.base_job import ContentType
from azure.quantum import Workspace, Job
from azure.quantum.job.base_job import DEFAULT_TIMEOUT
from azure.quantum.target.target import Target

if TYPE_CHECKING:
    from azure.quantum.optimization.problem import Problem

logger = logging.getLogger(__name__)

__all__ = [
    "RangeSchedule",
    "Solver",
]



class RangeSchedule:
    def __init__(self, schedule_type: str, initial: float, final: float):
        """Constructor of RangeSchedule for solvers.

        :param schedule_type:
            str, type of the RangeSchedule
            currently only supports 'linear' and 'geometric'.
        :param initial:
            float, initial value of RangeSchedule
        :param final:
            float, final value of the RangeSchedule
        """
        self.schedule_type = schedule_type
        self.initial = initial
        self.final = final
        if not (
            self.schedule_type == "linear" or self.schedule_type == "geometric"
        ):
            raise ValueError(
                '"schedule_type" must be "linear" or "geometric"!'
            )

class Solver(Target):
    def __init__(
        self,
        workspace: Workspace,
        name: str,
        provider_id: str,
        input_data_format="microsoft.qio.v2",
        output_data_format="microsoft.qio-results.v2",
        nested_params: bool = True,
        force_str_params: bool = False,
        params: dict = None,
        content_type : Optional[ContentType] = ContentType.json,
        **kwargs
    ):
        self.provider_id = provider_id
        self.nested_params = nested_params
        self.force_str_params = force_str_params
        self.params = {"params": {}} if nested_params else {}
        params = params or {}
        name = self._switch_name(name, **params)

        super().__init__(
            workspace=workspace,
            name=name,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            provider_id=provider_id,
            content_type = content_type,
            encoding="gzip",
            **kwargs
        )
 
        self._set_params(**params)
        self._check_params(**params)


    """Constants that define thresholds for submission warnings
    """
    SWEEPS_WARNING = 10000
    TIMEOUT_WARNING = 600

    def _switch_name(self, name, **params):
        return name

    def _check_params(self, **params):
        pass

    @staticmethod
    def _encode_input_data(data: "Problem") -> bytes:
        return data.to_blob()

    def submit(
        self, problem: Union[str, "Problem"]) -> Job:
        """Submits a job to execution to the associated
        Azure Quantum Workspace.

        :param problem:
            The Problem to solve. It can be an instance of a Problem,
            or the URL of an Azure Storage Blob where the serialized version
            of a Problem has been uploaded.
        """
        from azure.quantum.optimization import Problem
        if isinstance(problem, Problem):
            return super().submit(
                input_data=problem,
                name=problem.name,
                input_params=self.params,
                blob_name="inputData",
                content_type = problem.content_type
            )

        else:
            if hasattr(problem, "uploaded_blob_uri"):
                name = problem.name
                problem_uri = problem.uploaded_blob_uri

            elif isinstance(problem, str):
                name = "Optimization problem"
                problem_uri = problem
            
            else:
                raise ValueError("Cannot submit problem: should be of type str, Problem or have uploaded_blob_uri attribute.")

            # Create job from storage URI
            job = Job.from_storage_uri(
                workspace=self.workspace,
                name=name,
                target=self.name,
                input_data_uri=problem_uri,
                provider_id=self.provider_id,
                input_data_format=self.input_data_format,
                output_data_format=self.output_data_format,
                input_params=self.params,
                session_id=self.get_latest_session_id()
            )

        return job

    def optimize(self, problem: Union[str, "Problem"], timeout_secs: int=DEFAULT_TIMEOUT):
        """[Submits the Problem to the associated
            Azure Quantum Workspace and get the results.

        :param problem:
            The Problem to solve. It can be an instance of a Problem,
            or the URL of an Azure Storage Blob where the serialized version
            of a Problem has been uploaded.
        :type problem: Union[str, Problem]
        :param timeout_secs: Timeout in seconds, defaults to 300
        :type timeout_secs: int
        :return: Job results
        :rtype: dict
        """
        if not isinstance(problem, str):
            self.check_submission_warnings(problem)

        job = self.submit(problem)
        logger.info(f"Submitted job: '{job.id}'")

        return job.get_results(timeout_secs=timeout_secs)

    def set_one_param(self, name: str, value: Any):
        if value is not None:
            params = (
                self.params["params"] if self.nested_params else self.params
            )
            params[name] = str(value) if self.force_str_params else value

    def _set_params(self, **params):
        for param_name, param in params.items():
            self.set_one_param(param_name, param)

    def set_number_of_solutions(self, number_of_solutions: int):
        """Sets the number of solutions to return.

        Applies to all solvers.
        Default value is 1 if not supplied.

        :param number_of_solutions:
            Number of solutions to return. Must be a positive integer.
        """
        self.set_one_param("number_of_solutions", number_of_solutions)

    def check_submission_warnings(self, problem: "Problem"):
        # print a warning if we suspect the job
        # may take long based on its configured properties.
        is_large_problem = problem.is_large()
        if is_large_problem:
            if self.nested_params and "sweeps" in self.params["params"]:
                sweeps = int(self.params["params"]["sweeps"])
                # if problem is large and sweeps is large, warn.
                if sweeps >= Solver.SWEEPS_WARNING:
                    logger.warning(
                        f"There is a large problem submitted and \
                        a large number of sweeps ({sweeps}) configured. \
                        This submission could result in a long runtime."
                    )

        # do a timeout check if param-free, to warn
        # new users who accidentally set high timeout values.
        if self.nested_params and "timeout" in self.params["params"]:
            timeout = int(self.params["params"]["timeout"])
            if timeout >= Solver.TIMEOUT_WARNING:
                logger.warning(
                    f"A large timeout has been set for this submission \
                        ({timeout}). \
                        If this is intended, disregard this warning. \
                        Otherwise, consider cancelling the job \
                        and resubmitting with a lower timeout."
                )
    
    class ScheduleEvolution(Enum):
        INCREASING = 1
        DECREASING = 2

    def check_set_schedule(
        self,
        schedule_name: str,
        schedule_value: RangeSchedule,
        evolution: Optional[ScheduleEvolution] = None,
        lower_bound_exclusive: Optional[float] = None,
        lower_bound_inclusive: Optional[float] = None,
    ):
        """Set the parameter `schedule_name` from the value `schedule_value`
        and check that it is of type `RangeSchedule` and each end satisfies
        the specified bound.

        :param schedule_name:
            Name of the schedule parameter.
        :param schedule_value:
            Schedule value to be assigned and checked.
        :param evolution
            Expected schedule evolution (INCREASING or DECREASING)
        :lower_bound_exclusive:
            Exclusive lower bound for both ends of the schedule, optional.
        :lower_bound_inclusive:
            Inclusive lower bound for both ends of the schedule, optional.
        """
        if not (schedule_value is None):
            if not isinstance(schedule_value, RangeSchedule):
                raise ValueError(
                    f"{schedule_name} must be of type RangeSchedule; found"
                    f" type({schedule_name})={type(schedule_value).__name__}."
                )
            schedule_param = {
                "type": schedule_value.schedule_type,
                "initial": schedule_value.initial,
                "final": schedule_value.final,
            }
            if evolution is not None:
                if (evolution == self.ScheduleEvolution.INCREASING and
                        schedule_value.initial > schedule_value.final):
                    raise ValueError(
                            f"Schedule for {schedule_name} must be increasing;"
                            f" found {schedule_name}.initial"
                            f"={schedule_value.initial}"
                            f" > {schedule_value.final}"
                            f"={schedule_name}.final."
                    )
                if (evolution == self.ScheduleEvolution.DECREASING and
                        schedule_value.initial < schedule_value.final):
                    raise ValueError(
                            f"Schedule for {schedule_name} must be decreasing;"
                            f" found {schedule_name}.initial"
                            f"={schedule_value.initial}"
                            f" < {schedule_value.final}"
                            f"={schedule_name}.final."
                    )
            self.check_limit(
                    f"{schedule_name}.initial",
                    schedule_value.initial,
                    lower_bound_exclusive=lower_bound_exclusive,
                    lower_bound_inclusive=lower_bound_inclusive)
            self.check_limit(
                    f"{schedule_name}.final",
                    schedule_value.final,
                    lower_bound_exclusive=lower_bound_exclusive,
                    lower_bound_inclusive=lower_bound_inclusive)
            self.set_one_param(schedule_name, schedule_param)

    def check_set_positive_int(
            self,
            parameter_name: str,
            parameter_value: int):
        """Set the parameter `parameter_name` from the value `parameter_value`
        and check that it is a positive integer.

        :param parameter_name:
            Name of the parameter.
        :param parameter_value:
            Value to be assigned and checked.
        """
        if not (parameter_value is None):
            if not isinstance(parameter_value, int):
                raise ValueError(
                        f"{parameter_name} must be of type int; found"
                        f"type({parameter_name})"
                        f"={type(parameter_value).__name__}.")
            if parameter_value <= 0:
                raise ValueError(
                        f"{parameter_name} must be positive; found "
                        f"{parameter_name}={parameter_value}.")
            self.set_one_param(parameter_name, parameter_value)

    def check_set_float(
        self,
        parameter_name: str,
        parameter_value: float,
        lower_bound_exclusive: Optional[float] = None,
        lower_bound_inclusive: Optional[float] = None,
    ):
        """Set the parameter `parameter_name` from the value `parameter_value`
        and check that it has a float value satisfying bounds.

        :param parameter_name:
            Name of the parameter.
        :param parameter_value:
            Value to be assigned and checked.
        :lower_bound_exclusive:
            Exclusive lower bound to check parameter_value against, optional.
        :lower_bound_inclusive:
            Inclusive lower bound to check parameter_value against, optional.
        """
        if not (parameter_value is None):
            if not (isinstance(parameter_value, float) or
                    isinstance(parameter_value, int)):
                raise ValueError(f"{parameter_name} must be a float!")
            else:
                self.check_limit(
                        parameter_name=parameter_name,
                        parameter_value=parameter_value,
                        lower_bound_exclusive=lower_bound_exclusive,
                        lower_bound_inclusive=lower_bound_inclusive)
                self.set_one_param(parameter_name, parameter_value)

    def check_limit(
        self,
        parameter_name: str,
        parameter_value: Optional[float],
        lower_bound_exclusive: Optional[float] = None,
        lower_bound_inclusive: Optional[float] = None,
    ):
        """Check whether `parameter_value` satisfies a lower bound.

        :param parameter_name:
            Name of the parameter.
        :param parameter_value:
            Value to be checked.
        :lower_bound_exclusive:
            Exclusive lower bound to check parameter_value against, optional.
        :lower_bound_inclusive:
            Inclusive lower bound to check parameter_value against, optional.
        """
        if not (parameter_value is None):
            if (lower_bound_exclusive is not None and
                    parameter_value <= lower_bound_exclusive):
                raise ValueError(
                    f"{parameter_name} must be greater than "
                    f"{lower_bound_exclusive}; "
                    f"found {parameter_name}={parameter_value}."
                )
            if (lower_bound_inclusive is not None and
                    parameter_value < lower_bound_inclusive):
                raise ValueError(
                    f"{parameter_name} must be greater equal "
                    f"{lower_bound_inclusive}; found "
                    f"{parameter_name}={parameter_value}."
                )

